import enum


class Status(enum.IntEnum):
    NONE = 0

    DRIVING = 1

    STOPPED = INELIGIBLE = 2  # Ineligible to charge, but stopped
    AWAY_ELIGIBLE = 3      # Eligible for away charging due to length of stop
    HOME_ELIGIBLE = 4      # Eligible for home charging due to length of stop

    ARRIVED = enum.auto()
    IN_QUEUE = enum.auto()  # Currently in a charging queue
    CONNECTED = enum.auto()  # Currently connected to EVSE
    CONNECTED_V2G = enum.auto()  # Not implemented yet
    COMPLETED = enum.auto()
    DISCONNECTED = enum.auto()  # Disconnected



from typing import NamedTuple
import enum

# Unit conversion constants
_GAL_PER_L = 0.264172
_LB_PER_KG = 2.20462
_BTU_PER_MJ = 947.82
_KWH_PER_MJ = 0.2778

class Fuel(NamedTuple):
    """Energy information for a given fuel"""
    MJ_L: float # MegaJoules
    ci: float # gCO2e/MJ

    @classmethod
    def from_Btu(cls, btu: float, ci: float):
        return Fuel(
            MJ_L = btu / _BTU_PER_MJ,
            ci = ci * _BTU_PER_MJ
        )
    
    @classmethod
    def from_mix(cls, fuel_tuples: list):

        mj = 0
        ci = 0
        pct_total = 0.   
        for fuel, pct in fuel_tuples:
            if not isinstance(fuel, Fuel):
                raise TypeError('All fuels must be of type Fuel')
            
            mj += fuel.MJ_L * pct
            ci += fuel.ci * pct
            pct_total += pct
        
        if not (pct_total == 1.):
            raise ValueError('Percentages do not add up to 1')

        return Fuel(mj, ci)
    
    @property
    def MJ_gal(self):
        return self.MJ_L * _GAL_PER_L
    
    @property
    def kWh_gal(self):
        return (self.MJ_L * _GAL_PER_L) / _KWH_PER_MJ
    
    @property
    def kWh_L(self):
        return self.MJ_L / _KWH_PER_MJ
    
    @property
    def kgCO2_gal(self):
        return ((self.ci * self.MJ_L) * _GAL_PER_L) / 1000.0
    
    @property
    def kgCO2_L(self):
        return (self.ci * self.MJ_L) / 1000.0

"""Fuel Information

The following Fuel information is provided by the
California Energy Commission and Air Resources Board.

Energy values taken from CA Energy Commission:
https://www.energy.ca.gov/2007publications/CEC-600-2007-002/CEC-600-2007-002-D.PDF

CI Values from CARB's LCFS:
https://www.arb.ca.gov/fuels/lcfs/fuelpathways/pathwaytable.htm
"""

# "Pure" fuels
CARBOB = Fuel(31.8, 99.78) # CA's "regular" gasoline - Conv. Petrol
DieselULS = Fuel(35.5, 102.01) # Ultra-low-sulfur diesel - Conv. Petrol
CNG = Fuel(.038, 79.46) # Compressed Natural Gas - North American Conventional
E100_CAC = Fuel(21.2, 53.49) # Pure ethanol - California Corn (T1R-1516)
BD100_TUCO = Fuel(32.6, 30.15) # Pure biodiesel - Texan Used Cooking Oil (T1N-1735)

# Fuel Mixtures
CARBOB_E10 = Fuel.from_mix(
    [(CARBOB, .9),
    (E100_CAC, .1)]
)

CARBOB_E85 = Fuel.from_mix(
    [(CARBOB, .15),
    (E100_CAC, .85)]
)

DieselULS_BD20 = Fuel.from_mix(
    [(DieselULS, .8),
    (BD100_TUCO, .2)]
)



import numba as nb
import numpy as np

from .eligibility import EligibilityRules, Eligibility
from .bracket import Bracket, make_bracket


@nb.njit
def next_charging_opportunity(current_bracket: Bracket,
                              distance_array: np.array,
                              rules: EligibilityRules):

    if not (distance_array.ndim == 1):
        raise ValueError('Array of zero length supplied.')
    i = current_bracket.end
    out = None
    if current_bracket.end < distance_array.shape[0]:
        while i < distance_array.shape[0]:
            # if we're stopped:
            if distance_array[i] == 0.:
                bracket = make_bracket(distance_array, i, rules)

                if bracket.eligibility in {Eligibility.HOME, Eligibility.AWAY}:
                    out = bracket
                    break
            i += 1

    if bracket is None:
        out = Bracket(begin=distance_array.shape[0],
                      end=distance_array.shape[0])
    return out


@nb.njit
def distance_between_brackets(distance_array: np.array, b_start: Bracket,
                              b_end: Bracket):
    return distance_array[b_start.end:b_end.begin].sum()


@nb.njit
def is_last_bracket(distance_array: np.array, bkt: Bracket):
    return bkt.end >= distance_array.shape[0]


@nb.njit
def rewrite_deferred_trips(distance_array: np.array,
                           deferred_array: np.array,
                           ev_range: float, start_bracket: Bracket,
                           rules: EligibilityRules):

    # Only worry about doing this if:
    # 1: The vehicle is stopped
    # and 2: The ehicle is a BEV. PHEVs and ICEVs can leave any time.
    # and 3: The vehicle is about to leave, i.e.: the next interval
    # is a distance > 0

    current_bracket = start_bracket

    target_bracket = start_bracket

    while not is_last_bracket(distance_array, target_bracket):

        target_bracket = next_charging_opportunity(
            current_bracket,
            distance_array,
            rules
            )

        # Check to make sure the bracket a good one
        if target_bracket.begin < distance_array.shape[0]:
            dist = distance_between_brackets(
                distance_array,
                current_bracket,
                target_bracket)
            # If the BEV won't make it, rewrite!
            if dist > ev_range:
                x = current_bracket.end
                y = target_bracket.begin
                deferred_array[x:y] = distance_array[x:y]
                distance_array[x:y] = 0

                current_bracket = make_bracket(
                    distance_array,
                    current_bracket.index,
                    rules
                )
            # if the BEV can make it, breaks the loop!
            else:
                break

    return current_bracket



from typing import NamedTuple
import uuid

import pandas as pd
import numpy as np
import math

# Scenario information
from .scenario import Scenario
from .eligibility import EligibilityRules
from .fleet import Fleet
# Storage
from .datastorage import DatasetInfo, StorageInfo, DataHandler

# Objects
from .evse import EVSE, EVSEType
from .bank import Bank, QueueMode
from .status import Status

# The loop
from .loop import simulation_loop

__all__ = [
    'SimulationResult',
    'run_simulation'
]

class SimulationResult(NamedTuple):
    
    deferred: pd.DataFrame
    fuel_use: pd.DataFrame
    occupancy: pd.DataFrame
    demand: pd.DataFrame
    battery: pd.DataFrame
    scenario_id: str
    dataset_id: str
    dataset_path: str
    error: str = None
    
    def ok(self) -> bool:
        return not bool(self.error)


def run_simulation(ds: DatasetInfo, sc: Scenario, storage_info: StorageInfo):
    """Runs a simulation with a given scenario by
    downloading data, running the simulation, and uploading results.

    Args:
        sc (Scenario): The scenario structure to use.
    
    Returns:
        A `SimulationResult` with all the relevant energy data.
    """
    st = DataHandler(storage_info)

    try:
        # First pull down the data
        
        df = st.read_data(ds.obj_path)

        rows = len(df.index)

        interval_len = df.reset_index()['index'][:2].diff()[1].seconds / 60.0

        # Basic rules creation for now

        mask_df = pd.Series(
            np.full(len(df.index), False, dtype=np.dtype(bool)),
            index=df.index
        )

        for mask_rule in sc.home_mask_rules:
            if mask_rule.get('type') == 'time':
                begin = mask_rule.get('begin', '23:55')
                end = mask_rule.get('end', '00:00')
                mask_df.loc[df.between_time(begin, end).index] = True

        rules = EligibilityRules(
            away_threshold=round(sc.away_threshold_min/interval_len),
            home_threshold=round(sc.home_threshold_min/interval_len),
            mask=mask_df.values.astype(np.uint8)
        )

        # Create the fleet
        fleet = Fleet(df, sc.powertrains, sc.distribution, rules)

        # Get the banks up

        #away_banks = []
        #for bank_info in sc.away_banks:
        #    evse_list = []
        #    for evse_type in bank_info.get('evse'):
        #        evse_list.append(EVSE(evse_type))
        #    demand_a = np.zeros(rows, dtype=np.float32)
        #    occupancy_a = np.zeros(rows, dtype=np.uint8)
        #    bank = Bank(
        #        max_power=bank_info.get('max_power', 0.),
        #        capacity=bank_info.get('capacity', 0.),
        #        evse=evse_list,
        #        queue_probability=bank_info.get('probability', .2),
        #        queue_mode=bank_info.get('queue', QueueMode.DEFAULT),
        #        demand_profile=demand_a,
        #        occupancy_profile=occupancy_a,
        #        dynamic_size=True
        #    )
        #    away_banks.append(bank)

        away_banks = sc.make_away_banks(rows)
        
        #num_banks = 0
        #home_banks = []
        #for bank_info in sc.home_banks:
        #    evse_list = []
        #    for evse_type in bank_info.get('evse'):
        #        evse_list.append(EVSE(evse_type))
        #    bank = Bank(
        #        max_power=bank_info.get('max_power', 0.),
        #        capacity=bank_info.get('capacity', 0.),
        #        evse=evse_list,
        #        queue_probability=bank_info.get('probability', 1.),
        #        queue_mode=bank_info.get('queue', QueueMode.DEFAULT),
        #        demand_profile=np.zeros(rows, dtype=np.float32),
        #        occupancy_profile=np.zeros(rows, dtype=np.uint8),
        #        dynamic_size=False
        #    )
        #    home_banks.append(bank)
        #    num_banks += 1

        home_banks = sc.make_home_banks(fleet.size, rows)
        num_banks = len(home_banks)

        out_vehicles, out_banks = simulation_loop(
            fleet.vehicles,
            home_banks,
            away_banks,
            rules,
            interval_len,
            rows,
            sc.idle_load_kw
        )

        vehicle_ids = df.columns.values.tolist()
        fuel_df = pd.DataFrame(index=df.index)
        demand_df = pd.DataFrame(index=df.index)
        occupancy_df = pd.DataFrame(index=df.index)
        deferred_df = pd.DataFrame(index=df.index)
        battery_df = pd.DataFrame(index=df.index)

        for vehicle in out_vehicles:
            vid = str(vehicle.vid)
            fuel_df[vid] = vehicle.fuel_burned_a
            deferred_df[vid] = vehicle.deferred_a
            battery_df[vid] = vehicle.battery_a
        
        bank_i = 0
        for bank in out_banks:
            bank_col = 'bank_{}'.format(bank_i)
            demand_df[bank_col] = bank.demand_profile
            occupancy_df[bank_col] = (bank.occupancy_profile / bank.size) * 100.0 if bank.size > 0 else 0
            bank_i += 1
        if len(demand_df.columns) == 0:
            demand_df['bank_null'] = 0
        if len(occupancy_df.columns) == 0:
            demand_df['bank_null'] = 0
        
        obj_base = 'results/' + 'simulations/' + sc.run_id + '/'

        for fr, nm in zip([fuel_df, demand_df, battery_df, occupancy_df, deferred_df], ['fuel', 'demand', 'battery', 'occupancy', 'deferred']):
            
            st.upload_data(
                df=fr,
                obj_path=obj_base + nm + '/' +  ds.dataset_id,
                formats='json records csv',
            )
        
        

    except Exception as e:
        raise e

    finally:
        st.cleanup()

    return SimulationResult(
        scenario_id=sc.run_id,
        dataset_id=ds.dataset_id,
        dataset_path=ds.obj_path,
        deferred=deferred_df,
        fuel_use=fuel_df,
        occupancy=occupancy_df,
        demand=demand_df,
        battery=battery_df,
        error=None
    )    
    
    # return SimulationResult(errors=['Some error occurred.'])






import enum
from typing import NamedTuple

from .fuels import Fuel, CARBOB_E10
from .plug import DCPlug

__all__ = [
    'PType',
    'Powertrain'
]

class PType(enum.IntEnum):
    """Constants that refer to the type of drive.

    Valid values are:
    * `ICEV`: Conventional combustion and hybrids
            (i.e. Honda Civic, Toyota Prius)
    * `PHEV`: Plug-in hybrids (i.e. Chevrolet Volt)
    * `BEV`: Battery electric vehicles (i.e. Tesla Model 3)
    """
    ICEV = 0
    PHEV = 1
    BEV = 2


class Powertrain(NamedTuple):
    """A `NamedTuple` that contains information about the powertrain of a vehicle.

    Since the powertrain technical info is all that's needed for the
    simulation; make and model information don't really matter.

    However, a code (optional) is useful to keep track of what powertrain
    has been used.

    Attributes:
        id (str): A short code useful for looking up the object in a database.
        ice_eff (float): Fuel efficiency of the powertrain in km/L.
        ev_eff (float): Electric efficiency in km/kWh.
        batt_cap (float): Energy capacity of the battery in kWh.
        ac_power (float): Maximum input AC power in kW.
        dc_power (float): Maximum input DC power in kW.
        dc_plug (int): Connector type for DCFC.
        ptype (int): Type of powertrain as represented by a PType.
            Values may be of `PType.ICEV`, `PType.PHEV`, or `PType.BEV`.
        fuel (Fuel): The fuel that this powertrain uses. May be `None`
            if `ptype` is `PType.BEV`.
    """
    id: str = ''
    ice_eff: float = 32.
    ev_eff: float = 0.
    batt_cap: float = 0.
    ac_power: float = 0.
    dc_power: float = 0.
    dc_plug: int = DCPlug.NONE
    ptype: int = PType.ICEV
    fuel: Fuel = CARBOB_E10
    ice_alternator_eff = .21

    @property
    def pev(self) -> bool:
        """Returns `True` if the powertrain is a plug-in(BEV or PHEV)"""
        return self.ptype in {PType.BEV, PType.PHEV}

    @property
    def bev(self) -> bool:
        """Returns `True` if the vehicle is a BEV"""
        return self.ptype == PType.BEV

    @property
    def has_ice(self) -> bool:
        """Returns `True` there's a fuel-powered engine (ICEV or PHEV)"""
        return self.ptype in [PType.ICEV, PType.PHEV]

    @property
    def phev(self) -> bool:
        """Returns `True` if the the powertrain is a PHEV."""
        return self.ptype == PType.PHEV

    @property
    def icev(self) -> bool:
        """Returns `True` if the powertrain is a conventional ICEV/HEV"""
        return self.ptype == PType.ICEV

    @property
    def dc_capable(self) -> bool:
        """Returns `True` if the powertrain can accept AC connections."""
        return self.dc_power > 0

    @property
    def ac_capable(self) -> bool:
        """Returns `True` if the powertrain can accept DC connections."""
        return self.ac_power > 0

    def energy_at_soc(self, soc) -> float:
        """Returns how how much energy is in the battery at a given SoC."""
        return self.batt_cap * max(0.0, min(soc, 1.0))

    def idle_fuel_consumption(self, load_kwh, use_gal = True) -> float:
        if self.ice_eff > 0 and self.ice_alternator_eff > 0:
            nrg_needed = load_kwh / self.ice_alternator_eff
            if use_gal:
                return nrg_needed / self.fuel.kWh_gal
            else:
                return nrg_needed / self.fuel.kWh_L
        return 0




import evnrg

def main():
    """Prints a description of the package."""

    desc = """ EVNRG: An EV electrical demand and energy simulator.
    Version {}
    """.format(evnrg.__version__)

    print(desc)
    

if __name__ == "__main__":
    main()



from typing import NamedTuple

import numpy as np
import numba as nb

from .status import Status

__all__ = [
    'EligibilityRules',
    'ECode'
]

class EligibilityRules(NamedTuple):

    mask: np.array = np.empty((0, 0), dtype=np.bool)
    threshold_priority: int = Status.HOME_ELIGIBLE
    away_threshold: int = 0
    home_threshold: int = 0


@nb.njit
def check_mask(begin: int, end: int, col: int, rules: EligibilityRules):

    out = Status.INELIGIBLE
    if end < rules.mask.shape[0] and begin >= 0:
        if rules.mask.ndim == 2:
            out = max(rules.mask[begin:end, col].max(), 0)
        elif rules.mask.ndim == 1:
            out = max(rules.mask[begin:end].max(), 0)

    return out


class ECode(NamedTuple):
    """Small `NamedTuple` to indicate the stop's eligibility.
    
    Attributes:
        begin_index (int): Beginning index of the stop.
        end_index (int): Ending inde xof the stop.
            This is also the index of the beginning of the next
            trip.
        code (Status): The status of the a stop. Will be one of 
            `Status.INELIGIBLE`, `Status.HOME_ELIGIBLE`, or 
            `Status.AWAY_ELIGIBLE`. If the stop is not actually a stop,
            `Status.DRIVING` is used.

    """
    begin_index: int
    end_index: int
    code: int


@nb.jit
def stop_eligibility(distance_array: np.array,
                     begin: int,
                     vidx: int,
                     rules: EligibilityRules,
                     use_mask: bool = True):
    """Returns an `ECode` code indicating the charge eligibility."""

    status = Status.DRIVING
    end = begin
    if distance_array[begin] == 0:

        status = Status.INELIGIBLE

        alen = distance_array.shape[0]

        while end < alen and not (distance_array[end] > 0):
            end += 1

        length = end - begin
        # See if we've been stopped long enough to trigger
        # a charge opportunity
        home = False
        if rules.home_threshold > 0 and length > rules.home_threshold:
            home = True

        away = False
        if rules.away_threshold > 0 and length > rules.away_threshold:
            away = True

        if home and away:
            status = rules.threshold_priority
        elif home:
            status = Status.HOME_ELIGIBLE
        elif away:
            status = Status.AWAY_ELIGIBLE

        # Masks override any time-based rules
        # This allows for time-fencing and geo-fencing at the data
        # preparation level.
        if use_mask:
            charge_mask = check_mask(begin, end, vidx, rules)

            allowed_mask_values = (
                Status.HOME_ELIGIBLE,
                Status.AWAY_ELIGIBLE
            )

            if charge_mask in allowed_mask_values:
                status = charge_mask

    return ECode(
        begin_index=begin,
        end_index=end,
        code=status
    )



from dask.distributed import progress
from dask.distributed import Client

import dask
import os
import uuid
from typing import List
import math
import logging

from .scenario import Scenario
from .dataset import DatasetInfo
from .datastorage import StorageInfo
from .simulation import run_simulation, SimulationResult
from .numbaloop import run_simulation as nb_run_simulation
from .job_results import JobResults

class DaskJobRunner(object):

    __slots__ = (
        'scheduler_address',
        'client'
    )

    def __init__(self, scheduler_address = ''):

        self.scheduler_address = scheduler_address
        if not scheduler_address:
            self.client = Client()
            self.scheduler_address = 'localhost:8786'
        else:
            self.client = Client(self.scheduler_address)
    
    def run_simulations(self, scenarios: List[Scenario], datasets: List[DatasetInfo], 
                        storage_info: StorageInfo, use_numba = True):

        client = self.client

        logging.info('Running Scenarios...')

        if use_numba:
            print('Running simulations using Numba...')
        else:
            print('Running simulations using plain ol\' Python...')


        results = []

        for scenario in scenarios:
            for dataset in datasets:
                if use_numba:
                    
                    sim_result = dask.delayed(nb_run_simulation)(dataset, scenario, storage_info)
                    results.append(sim_result)
                else:
                    sim_result = dask.delayed(run_simulation)(dataset, scenario, storage_info)
                    results.append(sim_result)
          
        out = dask.compute(*results)
      
        return out
    
    



import os
import errno
import uuid
import re
import pathlib
from typing import NamedTuple

from libcloud.storage.base import Object, Container, StorageDriver
from libcloud.storage.types import (
    Provider,
    ContainerDoesNotExistError,
    ObjectDoesNotExistError
)
from libcloud.storage.providers import get_driver, DRIVERS
import appdirs
import pandas as pd

from .dataset import DatasetInfo

TMP_DIR = appdirs.user_cache_dir('evnrg')
CONFIG_DIR = appdirs.user_config_dir('evnrg')
LOCAL_DATA = appdirs.user_data_dir('evnrg')
LOCAL_STORAGE = os.path.join(LOCAL_DATA, 'local-object-storage')


# These are all aliases for other strings
PROVIDER_ALIASES = {
    # Amazon S3
    's3': Provider.S3,
    'amazon': Provider.S3,

    # Google Cloud Storage
    'gcs': Provider.GOOGLE_STORAGE,
    'google': Provider.GOOGLE_STORAGE,

    # Backblaze B2
    'b2': Provider.BACKBLAZE_B2,
    'backblaze': Provider.BACKBLAZE_B2,

    # Digital Ocean Spaces
    'digitalocean': Provider.DIGITALOCEAN_SPACES,
    'do': Provider.DIGITALOCEAN_SPACES,
    'spaces': Provider.DIGITALOCEAN_SPACES,

    # Microsoft Azure
    'microsoft': Provider.AZURE_BLOBS,
    'azure': Provider.AZURE_BLOBS,
}


class StorageInfo(NamedTuple):
    """A `NamedTuple` that tells the simulation how to handle data.
    
    Attributes:
        key (str): ID or login key for the cloud account. For example,
            Google Cloud Services uses an IAM email address, and
            can be found in the 'client_email' key of a generated
            JSON file. (See [Google Cloud's IAM documentation](https://cloud.google.com/iam/docs/)
            for more.) For local storage, the key is the base path
            that will be used while emulating cloud storage. Defaults
            to `~/.local/share/evnrg/local-object-storage`.
        secret (str): Secret key or passphrase for the cloud account. For
            Goggle Cloud, it will be in the 'private_key' field in the
            JSON file. Defaults to empty string ('').
        bucket (str): The name of the container (or bucket) to look for data
            and store it in. Defaults to 'evnrg-default'. For local storage,
            this folder will exist under the key: `<key>/<bucket>`.
        provider (str): The provider to use. Takes any provider string object
            that [Apache LibCloud](https://libcloud.readthedocs.io/en/latest/supported_providers.html#id195)
            takes. Additional provider shotcuts are provided as well:
            * Google Cloud Storage: 'gcs', 'google'
            * Amazon S3: 's3', 'amazon'
            * Backblaze B2: 'b2', 'backblaze'
            * Digital Ocean Spaces: 'do', 'spaces', 'digitalocean'
            * Microsoft Azure: 'azure', 'microsoft'
            Defaults to 'local'.
        cache_dir (str): The local cache directory for use when dumping
            dataframes for upload, and downloading dataframes for use.
            Defaults to `~/.cache/evnrg`.
        create_bucket (bool): Specifies if a bucket should be created if it
            does not exist. Defaults to `True`.
    """
    key: str = LOCAL_STORAGE
    secret: str = ''
    bucket: str = 'evnrg-default'
    provider: str = 'local'
    cache_dir: str = TMP_DIR
    create_bucket: bool = True


class UploadResult(NamedTuple):
    """Contains data about the uploaded object.

    Attributes:
        uid (str): The object's uuid.
        filetype (str): The object's filetype. Will generally be one of
            `'parquet'`, `'json'`, `'records'`, or `'csv'`.
        cache_path (str): The path to the temporary file.
        remote_path (str): The object's remote object name.
        obj (libcloud.storage.base.Object): The LibCloud storage object.
    """
    uid: str
    filetype: str
    cache_path: str
    remote_path: str
    obj: Object


class DataHandler(object):

    __slots__ = (
        'driver',
        'container',
        'cache_dir',
        'temp'
    )

    def __init__(self, storage_info: StorageInfo):

        if bool(storage_info.provider):
            provider_id = PROVIDER_ALIASES.get(
                storage_info.provider,
                storage_info.provider
            )
            if provider_id not in DRIVERS.keys():
                raise IndexError('Invalid object strage provider string.')

        self.driver = get_driver(provider_id)(
            storage_info.key,
            storage_info.secret
        )

        self.cache_dir = storage_info.cache_dir

        # Attempt to make the dir
        try:
            os.makedirs(self.cache_dir)
        except Exception as e:
            # Path already exists
            pass


        self.temp = []

        # Create the path if it doesn't exist
        pathlib.Path(self.cache_dir).mkdir(parents=True, exist_ok=True) 

        try:
            self.container = self.driver.get_container(
                container_name=storage_info.bucket
            )
        except ContainerDoesNotExistError:
            if storage_info.create_container:
                self.container = self.driver.create_container(
                        container_name=storage_info.bucket
                    )
            else:
                self.container = None

    @property
    def ready(self):
        return self.container is not None
    
    def use_bucket(self, bucket: str, create: False):
        try:
            self.container = self.driver.get_container(container_name=bucket)
            return True
        except ContainerDoesNotExistError:
            if create:
                try:
                    self.container = self.driver.create_container(
                        container_name=bucket
                    )
                    return True
                except Exception:
                    self.container = None
            else:
                self.container = None
        return False

    def cleanup(self):
        for fn in self.temp:
            if os.path.isfile(fn):
                os.remove(fn)

    def upload_data(self, df: pd.DataFrame, obj_path: str, 
                    formats: str = 'parquet',
                    keep_temp: bool = False, enable_cleanup: bool = True):

        uid = uuid.uuid4().hex

        fmts = re.findall(r'[\w]+', formats)

        local_path = os.path.join(self.cache_dir, uid)
        remote_base = obj_path.rstrip('/') + '/'

        results = []
        written = []

        if 'parquet' in fmts:
            local_parq = local_path + '.parquet'
            remote_parq = remote_base + 'parquet/' + uid + '.parquet'
            df.columns = df.columns.astype(str)
            df.to_parquet(
                local_parq,
                engine='fastparquet',
                compression='gzip'
            )
            written.append(local_parq)
            o = self.driver.upload_object(
                        local_parq,
                        self.container,
                        remote_parq
                    )
            if o:
                results.append(
                    UploadResult(
                        uid=uid,
                        filetype='parquet',
                        cache_path=local_parq,
                        remote_path=remote_parq,
                        obj=o
                    )
                )
        if 'json' in fmts:
            local_json = local_path + '.json'
            remote_json = remote_base + 'json/' + uid + '.json'
            df.to_json(
                local_json,
                orient='split',
                date_format='iso',
                date_unit='s'
            )
            written.append(local_json)
            o = self.driver.upload_object(
                        local_json,
                        self.container,
                        remote_json
                    )
            if o:
                results.append(
                    UploadResult(
                        uid=uid,
                        filetype='json',
                        cache_path=local_json,
                        remote_path=remote_json,
                        obj=o
                    )
                )
        if 'records' in fmts:
            local_records = local_path + '.records.json'
            remote_records = remote_base + 'records/' + uid + '.records.json'
            df.to_json(
                local_records,
                orient='records',
                date_format='iso',
                date_unit='s'
            )
            written.append(local_records)
            o = self.driver.upload_object(
                        local_records,
                        self.container,
                        remote_records
                    )
            if o:
                results.append(
                    UploadResult(
                        uid=uid,
                        filetype='records',
                        cache_path=local_records,
                        remote_path=remote_records,
                        obj=o
                    )
                )
        if 'csv' in fmts:
            local_csv = local_path + '.csv'
            remote_csv = remote_base + 'csv/' + uid + '.csv'
            df.to_csv(
                local_csv,
                date_format='%Y-%m-%dT%H:%M:%S'
            )
            written.append(local_csv)
            o = self.driver.upload_object(
                        local_csv,
                        self.container,
                        remote_csv
                    )
            if o:
                results.append(
                    UploadResult(
                        uid=uid,
                        filetype='csv',
                        cache_path=local_csv,
                        remote_path=remote_csv,
                        obj=o
                    )
                )

        # Delete temporary files
        if not keep_temp:
            for to_remove in written:
                if os.path.isfile(to_remove):
                    os.remove(to_remove)

        # Add to file list for cleanup
        if enable_cleanup:
            for to_remove in written:
                if os.path.isfile(to_remove):
                    self.temp.append(to_remove)

        return results

    def read_data(self, obj_name: str, fmt: str = 'parquet',
                  read_fn=None, arguments: dict = {}):
        o = None
        try:
            o = self.driver.get_object(self.container.name, obj_name)
        except ObjectDoesNotExistError:
            return None

        if o is None:
            return None

        fname = uuid.uuid4().hex + '-' + obj_name.split('/')[-1]

        tmp_name = os.path.join(self.cache_dir, fname)

        if self.driver.download_object(o, tmp_name, True, True):

            # retry 3 times
            tries = 0
            while not os.path.isfile(tmp_name) and tries < 3:
                self.driver.download_object(o, tmp_name, True, True)
                tries += 1


            if not os.path.isfile(tmp_name) and tries >= 3:
                return None
            
            self.temp.append(tmp_name)
            
            if fmt == 'parquet':
                # Default to fastparquet
                d_ = {'engine': 'fastparquet'}
                d_.update(arguments)
                return pd.read_parquet(tmp_name, **d_)

            elif fmt == 'csv':
                return pd.read_csv(tmp_name, **arguments)
            
            elif fmt == 'json':
                d_ = {'orient': 'split'}
                d_.update(arguments)
                return pd.read_json(tmp_name, **d_)
            
            elif fmt == 'json-records':
                d_ = {'orient': 'records'}
                d_.update(arguments)
                return pd.read_json(tmp_name, **d_)
            
            elif callable(read_fn):
                return read_fn(tmp_name, **arguments)
            
        return None
    
    def read_dataset(self, dataset_info: DatasetInfo, fmt: str = 'parquet',
                     read_fn=None, arguments: dict = {}):
        if not isinstance(dataset_info, DatasetInfo):
            raise TypeError('Must supply a DatasetInfo object!')

        return self.read_data(
            dataset_info.obj_path,
            fmt,
            read_fn,
            arguments
        ) 


class DataResource(object):

    __slots__ = ('handler', 'info')

    def __init__(self, storage_info: StorageInfo):
        self.info = storage_info

    def __enter__(self):

        self.handler = DataHandler(self.info)

        return self.handler

    def __exit__(self):

        self.handler.cleanup()


        

        

        



import numpy as np
import numba as nb
import pandas as pd
from typing import List
import random
from typing import NamedTuple

from .status import Status
from .eligibility import (
    EligibilityRules,
    ECode
)
from .eligibility import stop_eligibility
from .powertrain import Powertrain, PType
from .bracket import Bracket
from .plug import DCPlug

__all__ = [
    'Vehicle'
]

@nb.njit
def next_stop(distance_a: np.array, begin: int):
    i = begin
    length = distance_a.shape[0]
    d = distance_a[begin]
    if not (d == 0.):
        while i < length and not (d == 0):
            d = distance_a[i]
            i += 1
    return ECode(
        begin_index=begin,
        end_index=i,
        code=Status.DRIVING
    )
    

@nb.njit
def next_trip(distance_a: np.array, begin: int):
    i = begin
    length = distance_a.shape[0]
    while i < length and distance_a[i] == 0.:
        i += 1
    return i


@nb.njit
def next_charging_opportunity(distance_array: np.array,
                              begin: int,
                              vidx: int, rules: EligibilityRules):
    """Finds the next Bracket that's a viable charging opportunity.

    This is a JITted function.

    Args:
        current_bracket (Bracket): The `Bracket` to start searching from.
        distance_array (numpy.array): A 1-D Numpy distance array.
        vidx (int): The vehicle column index to use for mask lookup
        rules (EligibilityRules): The `EligibilityRules` object to use
    """

    if not (distance_array.ndim == 1):
        raise ValueError('Array of zero length supplied.')

    i = begin
    alen = distance_array.shape[0]
    code = ECode(alen, alen, Status.HOME_ELIGIBLE)
    if distance_array[begin] == 0. and not (distance_array[begin + 1] == 0.):
        while i < alen:
            if not (distance_array[i] == 0.):
                i += 1
            else:
                code = stop_eligibility(distance_array, i, vidx, rules)

                acceptable_codes = (
                    Status.HOME_ELIGIBLE,
                    Status.AWAY_ELIGIBLE
                )

                if code.code in acceptable_codes:
                    break

                else:
                    i = code.end_index

    return code


@nb.njit
def distance_between_brackets(distance_array: np.array, b_start: Bracket,
                              b_end: Bracket):
    """Returns the distance between two brackets

    Args:
        distance_array (numpy.array): A 1-D distance array
        b_start (Bracket): The `Bracket` to start from
        b_end (Bracket): The `Bracket` to end at

    Returns:
        A `float` of the total distance traveled in the bracket.
    """
    return distance_array[b_start.end:b_end.begin].sum()


@nb.njit
def is_last_bracket(distance_array: np.array, bkt: Bracket):
    """Returns if the bracket is the last bracket in thedistance array

    Args:
        distance_array (numpy.array): The 1-D `numpy.array` to test against
        bkt (Bracket): The `Bracket` to test

    Returns:
        `bool`
    """
    return bkt.end >= distance_array.shape[0]


@nb.vectorize(nopython=True)
def energy_req(d, eff, idle):
    out = 0    
    if d < 0:
        out =  idle
    elif d > 0:
        out =  d / eff
    return out  
#@nb.guvectorize([
#    nb.float32[:],
#    nb.float32,
#    nb.float32,
#    nb.float32,
#    nb.float32,
#    nb.float32,
#    nb.float32,
#    nb.float32[:],
#    nb.float32[:],
#    nb.float32[:]
#    ],
#    '(n),(),(),(),(),()->(n),(n),(n)', nopython=True)
#def energy_req(
#        distance_a: np.array,
#        interval_mins: float,
#        ev_batt: float,
#        ev_eff: float,
#        ev_idle_eff: float,
#        ice_eff: float,
#        ice_idle_eff: float,
#        out: np.array):
#    pev = bool(ev_eff > 0)
#    ice = bool(ice_eff > 0)
#    ELEC = 0
#    ICE = 1
#    SHORT = 2
#    batt = ev_batt
#    h = interval_mins / 60.0
#
#    for i in range(distance_a.shape[0]):
#        d = distance_a[i]
#
#        out[i, SHORT] = 0.
#
#        # Driving
#        if d >= 0:
#            if pev:
#                e_req = d / ev_eff
#                if (not ice) and (e_req > batt):
#                    out[i,SHORT] = e_req - batt
#                e_act = min(e_req, batt)
#                batt = batt - e_act
#                out[i, ELEC] = batt
#                
#                d = d - (e_act * ev_eff)
#            
#            if ice:
#                out[i, ICE] = d * ice_eff
#        
#        # Idling
#        elif d < 0:
#            h = interval_mins / 60.0
#            if pev:
#                e_req =  h * ev_idle_eff
#                if (not ice) and (e_req > batt):
#                    out[i, SHORT] = e_req - batt
#                e_act = min(e_req, batt)
#                batt = batt - e_act
#                out[i, ELEC] = batt                
#                h = h - (e_act / ev_eff)
#            
#            if ice:
#                out[i, ICE] = h * ice_idle_eff
#
class DriveDetail(NamedTuple):
    total_dist: float
    deferred: float
    driven: float
    net_elec_used: float
    fuel_used: float


@nb.njit
def drive_to_with_rw(
    distance_array: np.array,
    deferred_array: np.array,
    battery_array: np.array,
    fuel_array: np.array,
    begin: int,
    target: int,
    mins_per_interval: float,
    ev_eff: float,
    begin_ev_batt: float,
    max_ev_batt: float,
    ice_eff: float,
    ev_idle: float = 0.,
    ice_idle: float = 0.,
    range_added_per_interval: float = 0.,
    bev_buffer = 0. ):
    
    max_len = distance_array.shape[0]

    pev = bool(ev_eff > 0)
    ice = bool(ice_eff > 0)
    phev = pev and ice
    bev = pev and not ice

    if begin < max_len - 1:

        # set the beginning SOC
        batt = begin_ev_batt

        # If this is a bev
        if bev:
            batt = begin_ev_batt * (1 - bev_buffer)
        
        i = begin + 0
        run_again = True
        sz = target - begin
        z = np.zeros((sz, 3), dtype=np.float32)

        while i < target:

            z[:,:] = 0.
            
            energy_req(
                distance_array[i:target],
                mins_per_interval,
                batt,
                ev_eff,
                ev_idle,
                ice_eff,
                ice_idle,
                z
            )

            # If the required electricity exceeds our existing for a BEV, rewrite
            if bev and z[:,2].sum() > 0:
                # Forward through the current stop if it is one
                while (distance_array[i] == 0) and i < target:
                    i += 1
                # Rewrite until the next trip
                while (not (distance_array[i] == 0))  and i < target:
                    if range_added_per_interval > 0:
                        batt = min(batt + range_added_per_interval, max_ev_batt)
                    battery_array[i] = batt
                    deferred_array[i] = distance_array[i]
                    distance_array[i] = 0.
                    i += 1
            else:
                # Otherwise, we good and can break out
                break
        
        battery_array[i:target] = z[i:target, 0]
        fuel_array[i:target] = z[i:target, 1]
        total_driven = distance_array[i:target].sum()

        net_elec = battery_array[target - 1] - begin_ev_batt
        fuel_used = fuel_array[begin:target].sum()
        deferred = deferred_array[begin:target].sum()
        driven = distance_array[begin:target].sum()


    
    # Return true if a disconnection event happend
    if range_added_per_interval > 0 and total_driven > 0:
        return True
    return False


                

            



@nb.njit
def rewrite_deferred_trips(distance_array: np.array,
                           deferred_array: np.array,
                           interval_mins: float,
                           begin: int,
                           target: int,
                           begin_battery: float,
                           max_battery: float,
                           ev_eff: float,
                           idle_load_kw: float = 0.,
                           energy_added_per_interval: float = 0.,
                           soc_buffer: float = 0.):
    """Searches for next eligible stop for BEVs, and defers trips as necessary.

    If a BEV's range can't achieve the distance, the trips are deferred until
    the distance is less than the range available. Additionally, This allows
    vehicles in a queue to stay in it, or vehicles to continue charging.

    Note:
        This function modifies the distance and deferred arrays in-place.

    Args:
        distance_array (numpy.array): The distance array to use
        deferred_array (numpy.array): The array that will be used to record
            deferred distance
        begin (int): The index to begin the operation at
        vidx (int): Vehicle column index (fleet_id)
        begin_status (int): The `Status` at the beginning index.
        begin_ev_range (float): Current range remaining in vehicle's battery
        max_ev_range (float): Maximum range the vehicle's battery can have
        rules (EligibilityRules): the `EligibilityRules` to use
        range_added_per_interval (:obj: `float`, optional): Allows range to be
            added if the current state is `Status.CONNECTED`.

    Returns:
        A revised `Status`.

    """

    # Only worry about doing this if:
    # 1: The vehicle is stopped
    # and 2: The ehicle is a BEV. PHEVs and ICEVs can leave any time.
    # and 3: The vehicle is about to leave, i.e.: the next interval
    # is a distance > 0

    max_len = distance_array.shape[0]

    if begin < max_len - 1:

        batt = begin_battery

        soc_coe = 1.0 - soc_buffer
        
        i = begin

        while i < target:
            # is the index advanced to a trip?

            length = target - i
            #short = np.zeroes(length, dtype=np.float32)
            #elec = np.zeroes(length, dtype=np.float32)
            #fuel = np.zeroes(length, dtype=np.float32)

            required_energy = energy_req(
                    distance_array[i:target],
                    (interval_mins / 60.0) * idle_load_kw,
                    ev_eff
                ).sum()

            if batt * soc_coe >= required_energy:
                break

            if not (distance_array[i] == 0):
                j = i
                # Find the end of the trip
                while not (distance_array[j] ==  0):
                    # Rewrite trips, but not idles
                    if distance_array[j] > 0:
                        deferred_array[j] = distance_array[j]
                        distance_array[j] = 0.0
                    if energy_added_per_interval > 0 and batt + energy_added_per_interval < max_battery:
                        batt += energy_added_per_interval
                    j += 1
                i = j
            else:
                # Iterate through a non-eligible stop
                i += 1
        
    return distance_array, deferred_array


@nb.njit
def drive_to_next_stop(begin: int, distance_a: np.array, interval_min: float, fuel_a: np.array,
                       battery_a: np.array, battery_start: float,
                       ev_efficiency: float, fuel_efficiency: float, 
                       fuel_alt_eff: float, fuel_kwh_per: float,
                       idle_load_kw: float):
    """Efficiently drives a vehicle through a drive bracket and stores energy data.

    Args:
        distance_a (numpy.array): 1-D Numpy distance array
        fuel_a (numpy.array); 1-D Numpy array to record fuel use
        battery_a (numpy.array): 1-D Numpy array that holds battery state
        battery_start (float): The starting battery energy for the vehicle
        ev_efficiency (float): The efficiency of electric operation in km/kWh.
            For ICEVs, this should be zero.
        fuel_efficiency (float): The efficiency of combustion operation
            in km/L. For BEVs, this should be zero.

    Returns:
        An `int` indicating the number of intervals traveled.

    Raises:
        ValueError: if the EV distance and Ice distance do not add up to the
            interval's distance (meaning the EV could not make it).

    """
    error_ = .001
    # Double check we're driving
    i = begin
    interval_hrs = interval_min / 60.0
    use_fuel_idle = bool(fuel_alt_eff > 0 and fuel_kwh_per > 0)
    if not (distance_a[i] == 0.):
        max_len = distance_a.shape[0]
        while (i < max_len) and not (distance_a[i] == 0.):

            distance = distance_a[i]
            e_dist = 0.
            e_idle = idle_load_kw * interval_hrs
            fuel_dist = 0.
            

            # Check if we have a pev
            if ev_efficiency > 0.:
                # Do this to make sure we don't try to access index of -1
                battery_state = battery_start
                if i > 0:
                    battery_state = battery_a[i - 1]

                # Only continue if we actually have battery power left
                if battery_state > 0.:

                    # Process battery
                    battery_required = 0.

                    # Driving
                    if distance >= 0:
                        battery_required = distance / ev_efficiency
                    
                    # Idle stops
                    elif distance < 0:
                        battery_required = e_idle
                    
                    battery_used = min(battery_state, battery_required)
                    battery_a[i] = battery_state - battery_used

                    # Driving
                    if distance >= 0:
                        e_dist = round(battery_used * ev_efficiency, 4)
                    # Idle stops
                    elif idle_load_kw > 0:
                        e_idle = round(e_idle - battery_used, 4)

            if fuel_efficiency > 0.:

                fuel_used = 0
                
                if distance >= 0:
                    fuel_used = (distance - e_dist) / fuel_efficiency

                elif e_idle > 0 and use_fuel_idle:
                    fuel_used = e_idle / (fuel_alt_eff * fuel_kwh_per)

                fuel_a[i] = fuel_used

            i += 1

    return i


class Vehicle(object):
    """The primary driving logic object.

    The `Vehicle` object keeps track of all data and operations specific
    to an individual vehicle. Many of the operations are sped up through
    external JITed functions.

    Attributes:
        fleet_id (int): The column index of the vehicle for use when
            referencing 2-D array or positions in a `Fleet` object.
        powertrain (Powertrain): The `'Powertrain` used for this vehicle's
            energy calculations.
        max-soc (float): The maximum state of charge (0.0 to 1.0) this
            vehicle's battery may reach. This will be handy for modeling
            battery degradation. For now, it defaults to 1.
        evse_power (float): The effective power level in kW that this vehicle
            is connected to. If the vehicle is not connected to EVSE, this
            value is 0.
        distance_a (numpy.array): A 1-D array of type `numpy.float32` that
            contains the interval distance data. Unless deferring and
            rewriting trips, this array should generally be treated as
            read-only.
        fuel_burned_a (numpy.array): A 1-D array of type `numpy.float32` that
            holds fuel consumption data.
        deferred_a (numpy.array): A 1-D array of type `numpy.float32` that
            holds deferred distance. When BEVs have to defer trips, the
            distances are copied to this array before the window in
            `distance_a` is set to zero. This offers useful data about trip
            completion by BEVs.
        battery_a (numpy.array): A 1-D array of type 'numpy.float32` that
            reflects the battery energy of the vehicle over time.
        idx (int): The internal index for the `Vehicle` object.
        begin_energy (float): The energy the vehicle starts with. This value
            is referenced in battery operations when `idx` is 0.
        status (int): indicates the current status of the vehicle. The value
            is held by an `int`, but is generally set by referencing the
            `Status` object (which is an `enum.IntEnum`). Do not try to
            evaluate this value on its own. Instead use `Status`
            (e.g. ``vehicle.status == Status.CONNECTED``

    Args:
        fleet_index (int): The vehicle column index from a `Fleet` object.
        ptrain (Powertrain): The `Powertrain` object that will be used with
            this vehicle.
        distance (numpy.array): A 1-D `numpy.array` of type `numpy.float32`
            that holds the interval distance data for this vehicle.
        rules (EligibilityRules): The `EligibilityRules` that will be used
            to evaluate the initial charging eligibility if `start_soc` is
            less than 1.0.
        start_soc (float): The initial state of charge of the vehicle. Defaults
            to 1.0 (100%).
    """

    __slots__ = (
        'vid',
        'fleet_id',
        'powertrain',
        'max_soc',
        'evse_power',
        'distance_a',
        'fuel_burned_a',
        'deferred_a',
        'battery_a',
        'idx',
        # 'bracket',
        'begin_energy',
        'status',
        'soc_buffer'
    )

    def __init__(self, fleet_index: int, 
                 ptrain: Powertrain, distance: np.array,
                 rules: EligibilityRules, start_soc: float = 1.,
                 soc_buffer: float = 0, vid: str = ''):
        
        self.vid = vid
        self.begin_energy = ptrain.energy_at_soc(start_soc)
        self.fleet_id = fleet_index
        self.powertrain = ptrain
        self.battery_a = np.zeros(distance.shape, dtype=np.float32)
        self.max_soc = 1 if ptrain.pev else 0
        self.evse_power = 0.
        self.distance_a = distance
        self.fuel_burned_a = np.zeros(len(distance), dtype=np.float32)
        self.soc_buffer = float(soc_buffer)
        self.deferred_a = np.zeros(len(distance), dtype=np.float32)

        self.idx = 0

        self.status = Status.DRIVING
        if distance[0] == 0.:
            if start_soc < 1.:
                self.status = stop_eligibility(0, distance, fleet_index, rules)
            else:
                self.status = Status.STOPPED

        # self.bracket = Bracket(0, fleet_index, distance_array, rules)

    @property
    def battery_state(self):
        """Provides the "incoming" battery state (from the previous index).

        Returns:
            A `float` representing the battery energy.
        """
        if self.idx > 0:
            return self.battery_a[self.idx - 1]
        return self.begin_energy

    @property
    def intervals(self):
        """The length of the distance array.
        
        Returns:
            A length of type `int`.
        
        Raises:
            `AssertionError` if the length is zero.
        """
        d = self.distance_a.shape[0]

        assert d > 0, \
            'Cannot use a zero-length drive profile.'
        return d

    @property
    def is_driving(self):
        """Returns a `bool` indicating if the vehicle is driving."""
        return self.distance_a[self.idx] > 0

    @property
    def is_stopped(self):
        """Returns the opposite of `is_driving`."""
        return not self.is_driving

    @property
    def is_new_stop(self):
        out = False
        if self.idx > 0:
            out = self.is_stopped and self.distance_a[self.idx - 1] > 0
        return out

    @property
    def will_drive_next(self):
        """Indicates whether the next interval will start a trip."""
        return self.is_stopped and self.distance_a[self.idx + 1] > 0

    @property
    def elapsed_distance(self):
        """Returns the elapsed distance for this simulation"""
        return np.sum(self.distance_a[:self.idx])

    @property
    def total_distance(self):
        """Returns the total distance for the simuilation.
        Note:
            This value will change if trips are deferred!
        """
        return np.sum(self.distance_a)
    
    @property
    def progress(self):
        """Returns how far through its distance array the vehicle is.
        The value is expressed as a `float` percentage wher 1.0 is 100%.
        """
        return (self.idx + 1) / self.intervals

    @property
    def distance(self):
        return self.distance_a[self.idx]

    @property
    def soc(self):
        """Returns the battery state of charge as a percentage."""
        out = 0
        if self.powertrain.batt_cap > 0:
            out = self.battery_a[self.idx - 1] / self.powertrain.batt_cap
        return out

    @property
    def ev_range(self):
        """Returns the range (in km) the battery has remaining."""
        return self.battery_a[self.idx - 1] * self.powertrain.ev_eff

    @property
    def max_ev_range(self):
        """The electric range the vehicle would have on a fully-charged
        battery.
        """
        return self.powertrain.batt_cap * self.max_soc * self.powertrain.ev_eff

    @property
    def evse_connected(self) -> bool:
        """Indicates if the attribute `evse_power` is greater than zero.
        A value of `True` indicates that an EVSE is connected to the vehicle.
        """
        return self.evse_power > 0.

    def index_sync(self, index: int):
        """Reports back how many intervals ahead or behind the vehicle is.

        Args:
            index (int): The index to check against

        Returns:
            An `int` indicating the differnce between `index` and the
                vehicle's index. A negative number indicates the vehicle is
                ahead. A positive number indicates it is behind. Zero
                indicates the indexes are in sync.
        """
        return index - self.idx

    def increment_index(self, index: int):
        """Advances the vehicle's index, but only if ithe vehicle is behind.

        Args:
            index (int): The index to attempt to advance to.

        Returns:
            An `int` indicating the number intervals behind or forward
                the vehicle's internal index is.

        """
        if index > self.idx:
            self.idx += 1
        return index - self.idx

    def set_battery(self, val: float, previous: bool = True):
        """Sets the battery energy.

        Args:
            val (float): The value to set the battery to.
                Nomalizes to be between 0 and the battery's max.
            previous (:obj: `bool`, optional): Whether to set the
                previous index's value. Defaults to `True`.
                If the current index is 0, then sets the beginning
                value for the vehicle.
        """
        true_val = max(0, min(self.powertrain.batt_cap, val))
        i = self.idx - 1 if previous else self.idx
        if i >= 0:
            self.battery_a[i] = true_val
        else:
            self.begin_energy = true_val

    def delta_battery(self, delta: float):
        """Sets the current index's battery state.

        The battery will be normalized to zero or the battery max.

        Args:
            delta (float): The change in the battery.
        Returns:
            A `float` representing the current battery energy.
        """
        val = self.battery_state + delta
        true_val = max(0., min(self.powertrain.batt_cap, val))
        self.battery_a[self.idx] = true_val
        return true_val

    def evse_compatible(self, dcfc: bool, plugs=None):

        if dcfc:
            if self.powertrain.dc_capable:
                if self.powertrain.dc_plug in plugs:
                    return True
            return False
        return True


    def connect_evse(self, power, dcfc, plugs=None):
        """Attempts to connect EVSE to the vehicle.

        Checks to see if the connection is compatible.

        Args:
            power (float): The input power from the EVSE
            dcfc (bool): Whether the input is DC.
            plugs (:obj: iterable, optional): An iterable
                containing the DC connectors available at
                this EVSE. Defaults to None.
                This is only needed for DC connections. AC
                connections are assumed to be J1772.

        Returns:
            `False` if connection fails, or a `float` representing
            the effective power level of the conenction (in kW).
        """

        if dcfc and self.powertrain.dc_power > 0.:
            # Check for DCFC plug compatibility
            if self.powertrain.dc_plug in plugs:
                self.evse_power = min(self.powertrain.dc_power, power)
                self.status = Status.CONNECTED
                return self.evse_power

        elif not dcfc and self.powertrain.ac_power > 0.:
            self.evse_power = min(self.powertrain.ac_power, power)
            self.status = Status.CONNECTED
            return self.evse_power

        return False

    def disconnect_evse(self):
        """Disconnects the EVSE and sets the power level to zero."""
        self.status = Status.DISCONNECTED
        self.evse_power = 0.

    def charge_battery(self, energy: float):
        """Charges the battery.

        Essentially a wrapper for `Vehicle.delta_battery()`,
        except that this function checks to ensure there
        is EVSE connected.

        Args:
            energy (float): The energy to add

        Returns:
            A `float` representing net energy added.
                This adjusts for the fact the battery may
                be nearly full.
        """

        if self.powertrain.pev and self.evse_connected:
            new_state = self.delta_battery(energy)
            return self.battery_state - new_state
        return False
    
    def drive(self, interval_min, idle_kw):

        idle_req = idle_kw * (interval_min / 60.0)
        d = self.distance_a[self.idx]

        idle = bool(d < 0)

        if self.powertrain.pev:
            req = 0
            if d < 0:
                req = self.delta_battery(-1 * idle_req) 
                idle_req = idle_req + req
            elif d > 0:
                req = self.delta_battery(-1 * (d / self.powertrain.ev_eff))
                d = d + (req * self.powertrain.ev_eff)
            elif not self.evse_connected:
                self.battery_a[self.idx] = self.battery_state

        fuel = 0
        if self.powertrain.ice_eff > 0:
            if idle and idle_req > 0:
                fuel = self.powertrain.idle_fuel_consumption(idle_req)
            elif d > 0:
                fuel = d / self.powertrain.ice_eff
        self.fuel_burned_a[self.idx] = fuel

    def attempt_defer_trips(self, rules: EligibilityRules,
                            min_per_interval: float, idle_load_kw: float = 0.):
        
        if self.idx + 1 < self.distance_a.shape[0]:

            if (self.powertrain.bev and self.will_drive_next):

                acceptable_codes = (
                        Status.HOME_ELIGIBLE
                )

                max_len = self.distance_a.shape[0]
                
                target_index = next_trip(self.distance_a, self.idx)
                
                el_stop = None
                while el_stop is None and target_index < max_len:

                    stop = next_stop(self.distance_a, target_index)

                    code = stop_eligibility(
                        self.distance_a,
                        stop.end_index,
                        self.fleet_id,
                        rules,
                        use_mask=False
                    )

                    if code.begin_index == code.end_index:
                        target_index = max_len
                    elif code.code == Status.HOME_ELIGIBLE:
                        el_stop = code
                    else:
                        target_index = code.end_index
                
                a, b = rewrite_deferred_trips(
                    self.distance_a,
                    self.deferred_a,
                    min_per_interval,
                    self.idx,
                    target_index,
                    self.battery_state,
                    self.powertrain.energy_at_soc(0.9),
                    idle_load_kw,
                    self.evse_power * (min_per_interval/60.0),
                    self.soc_buffer
                )

                self.distance_a = a
                self.deferred_a = b

                # Recheck eligibility
                if self.powertrain.pev and self.status == Status.INELIGIBLE:
                    new_code = stop_eligibility(
                            self.distance_a,
                            self.idx,
                            self.fleet_id,
                            rules,
                            use_mask=False
                        )
                    self.status = new_code.code

    def advance_index(self, rules: EligibilityRules, min_per_interval: float, idle_load_kw: float = 0.):

        max_len = self.distance_a.shape[0]
        if self.idx < max_len:

            running = not (self.distance_a[self.idx] == 0.)
            run_next = False
            if self.idx < max_len - 1:
                run_next = not (self.distance_a[self.idx + 1] == 0.)
            
            self.drive(min_per_interval, idle_load_kw)                    
            
            self.idx += 1

            code = stop_eligibility(
                        self.distance_a,
                        self.idx,
                        self.fleet_id,
                        rules
                    )
            self.status = code.code

                #code: ECode
                #code = next_stop(self.distance_a, self.idx)
                #try:

                #    stop_index = drive_to_next_stop(
                #        self.idx,
                #        self.distance_a[:],
                #        min_per_interval,
                #        self.fuel_burned_a[:],
                #        self.battery_a[:],
                #        self.battery_state,
                #        self.powertrain.ev_eff,
                #        self.powertrain.ice_eff,
                #        self.powertrain.ice_alternator_eff,
                #        self.powertrain.fuel.kWh_gal,
                #        idle_load_kw
                #    )
                
                #except AssertionError:
                #    pass

                #self.idx = stop_index
                

            #elif run_next:
            #    self.attempt_defer_trips(rules, min_per_interval)
            #    self.idx += 1
            #    if self.idx < max_len:
            #        self.battery_a[self.idx] = self.battery_a[self.idx - 1]
            #else:
            #    self.idx += 1
            #    if self.idx < max_len:
            #        self.battery_a[self.idx] = self.battery_a[self.idx - 1]



import random
import math
from typing import NamedTuple
import uuid

import numba as nb
import numpy as np
import pandas as pd


from .scenario import Scenario
from .datastorage import DatasetInfo, StorageInfo, DataHandler
from .evse import EVSE, EVSEType
from .bank import Bank
from .powertrain import Powertrain, PType
from .plug import DCPlug



ICEV = 0
PHEV = 1
BEV = 2



EVSE_POWER_MAX = 0
EVSE_BANK_POWER_MAX = 1
EVSE_POWER = 2
EVSE_MAX_SOC = 3
EVSE_PROBABILITY = 4
EVSE_DC = 5
DCPLUG_CHADEMO = 6 # Index for EVSE
DCPLUG_COMBO = 7
DCPLUG_TESLA = 8
DCPLUG_NONE = 0 # None Type

evse_ = np.dtype([
    ('bank_id', np.uint64, 1),
    ('power_max', np.float32, 1),
    ('bank_power_max', np.float32, 1),
    ('power', np.float32, 1),
    ('max_soc', np.float32, 1),
    ('probability', np.float32, 1),
    ('dc', np.bool_, 1),
    ('plug_chademo', np.bool_, 1),
    ('plug_combo', np.bool_, 1),
    ('plug_tesla', np.bool_, 1)
])

nb_evse_ = nb.from_dtype(evse_)

def make_evse_banks(evse_banks: list, fleet_size: int):

    n_evse = 0

    bank_ids = []
    power_max = []
    bank_power_max = []
    max_soc = []
    probability = []
    dc = []
    plug_chademo = []
    plug_combo = []
    plug_tesla = []  

    bank_id = 0
    for bank in evse_banks:
        
        bank_running_power = 0
        n_in_bank = 0
        for e, min_, max_, prop_ in bank['evse']:
            e: EVSEType
            copies = int(fleet_size * prop_)
            if max_ > 0:
                copies = int(min(max_, copies))
            copies = int(max(min_, copies))

            for i in range(copies):
                bank_ids.append(bank_id),
                power_max.append(e.max_power)
                
                max_soc.append(e.max_soc)
                probability.append(bank.get('probability', 1.))
                dc.append(e.dc)
                plug_chademo.append(bool(DCPlug.CHADEMO in e.dc_plugs))
                plug_combo.append(bool(DCPlug.COMBO in e.dc_plugs))
                plug_tesla.append(bool(DCPlug.TESLA in e.dc_plugs))
                n_evse += 1
                bank_running_power += e.max_power
                n_in_bank += 1
        bank_id += 1
        bank_power_max_val = bank,get('max_power', bank_running_power)
        bank_power_max.extend([bank_power_max_val for i in range(n_in_bank)])
    
    a = np.empty(n_evse, dtype=evse_)

    a[:]['bank_id'] = bank_ids
    a[:]['power'] = 0.
    a[:]['power_max'] = power_max
    a[:]['bank_power_max'] = bank_power_max
    a[:]['max_soc'] = max_soc
    a[:]['probability'] = probability
    a[:]['dc'] = dc
    a[:]['plug_chademo'] = plug_chademo
    a[:]['plug_combo'] = plug_combo
    a[:]['plug_tesla'] = plug_tesla

    return a

@nb.njit(cache=True)
def throttle_bank(bank: np.array, max_factor: float = .9):
    bank_total = bank[:, EVSE_POWER].sum()
    bank_max = bank[0, EVSE_BANK_POWER_MAX]

    scale_up = ((bank_max * max_factor) - bank_total) / bank_total
    scale_down = ((max_factor * bank_max) / bank_total)

    for i in range(bank.shape[0]):
        if bank[i, EVSE_POWER] > 0:
            if bank_total > bank_max:
                p = bank[i, EVSE_POWER] * scale_down
                bank[i, EVSE_POWER] = p
            elif bank_total < bank_max:
                p = bank[i, EVSE_POWER] + (bank[i, EVSE_POWER] * scale_up)
                bank[i, EVSE_POWER] = min(p, bank[i, EVSE_POWER_MAX])
    
    return bank

# Indexes in a vehicle
VEHICLE_ATTRIBUTE_LENGTH = 9

TYPE = 0
ICE_EFF = 1
ICE_GAL_PER_KWH = 2
EV_EFF = 3
EV_MAX_BATT = 4
AC_MAX = 5
DC_MAX = 6
DCPLUG_TYPE = 7
CONNECTED_EVSE_ID = 8

vehicle_ = np.dtype([
    ('type', np.int32, 1),
    ('ice_eff', np.float32, 1),
    ('ice_gal_kwh', np.float32, 1),
    ('ev_eff', np.float32, 1),
    ('ev_max_batt', np.float32, 1),
    ('ac_max', np.float32, 1),
    ('dc_max', np.float32, 1),
    ('dc_plug', np.int32, 1),
    ('home_evse_id', np.int64, 1),
    ('away_evse_id', np.int64, 1)
])

nb_vehicle_ = nb.from_dtype(vehicle_)

def make_fleet(
    vtypes: list,
    ice_effs: list,
    ice_gal_kwhs: list,
    ev_effs: list,
    ev_max_batts: list,
    ac_maxes: list, 
    dc_maxes: list,
    dcplugs: list):

    n = len(vtypes)
    a = np.empty(n, dtype=vehicle_)


    a[:]['type'] = vtypes
    a[:]['ice_eff'] = ice_effs
    a[:]['ice_gal_kwh'] = ice_gal_kwhs
    a[:]['ev_eff'] = ev_effs
    a[:]['ev_max_batt'] = ev_max_batts
    a[:]['ac_max'] = ac_maxes
    a[:]['dc_max'] = dc_maxes
    a[:]['dc_plug'] = dcplugs
    a[:]['home_evse_id'] = -1
    a[:]['away_evse_id'] = -1
    
    return a

def fleet_from_df(df: pd.DataFrame, powertrains: list,
                 probabilities: list):
        
        
    vnames = df.columns

    size = len(df.columns)

    pt_assignments = None

    if isinstance(powertrains, list):
        for pt in powertrains:
            if not isinstance(pt, Powertrain):
                raise TypeError()

        trains = powertrains
        num_trains = len(powertrains)
        num_probs = len(probabilities)
        probs = probabilities
        if not probabilities:
            probs = [float(1.0 / num_trains) for i in range(num_trains)]
        elif num_probs < num_trains:
            current_sum = sum(probs)
            if current_sum < 1:
                remaining = num_trains - num_probs
                distribution = (1 - current_sum) / remaining
                probs.extend([distribution for i in range(remaining)])
            else:
                trains = powertrains[:num_probs]
        elif num_trains < num_probs:
            probs = probabilities[:num_trains]

        pt_assignments = np.random.choice(
            range(num_trains),
            size=size,
            replace=True,
            p=np.array(probs)
        )

    vtypes = []
    ice_effs = []
    ice_gal_kwhs = []
    ev_effs = []
    ev_max_batts = []
    ac_maxes = [] 
    dc_maxes = []
    dcplugs = []

    for i in pt_assignments:
        pt = powertrains[i]
        pt: Powertrain
            
        vtypes.append(int(pt.ptype))
        ice_effs.append(pt.ice_eff)
        ice_gal_kwhs.append(pt.idle_fuel_consumption(1))
        ev_effs.append(pt.ev_eff)
        ev_max_batts.append(pt.batt_cap)
        ac_maxes.append(pt.ac_power)
        dc_maxes.append(pt.dc_power)
        dcplugs.append(pt.dc_plug)
        
    fleet = make_fleet(
        vtypes,
        ice_effs,
        ice_gal_kwhs,
        ev_effs,
        ev_max_batts,
        ac_maxes,
        dc_maxes,
        dcplugs
    )

    distance = np.array(df.values, dtype=np.float32)
    
    return distance, fleet

@nb.njit(cache=True)
def bank_enqueue(idx, vid, soc, fleet, queue, queue_soc):

    if fleet[vid]['type'] in (PHEV, BEV):
        if soc < 1:
            if queue_soc:
                queue[vid] = soc
            else:
                queue[vid] = idx
    
    return queue

@nb.njit(cache=True)
def pop_low_score(queue):
    idx = -1
    for i in range(queue.shape[0]):
        x = queue[i]
        if not np.isnan(x):
            if idx == -1:
                idx = i
            elif x < queue[idx]:
                idx = i
    if (not (np.isnan(idx))) and (idx >= 0):
        queue[nb.int64(idx)] = np.nan
    return idx

@nb.njit(cache=True)
def bank_dequeue(queue, idx):
    queue[idx] = np.nan
    return queue

@nb.njit(cache=True)
def get_soc(vid, fleet, battery_nrg):
    out = 0.
    if fleet[vid]['ev_max_batt'] > 0:
        out = battery_nrg / fleet[vid]['ev_max_batt']
    return out

@nb.njit(cache=True)
def connect_evse(vid, soc, fleet, bank, away_bank = False):

    evse_ids = range(bank.shape[0])
    connected_power = 0

    for i in evse_ids:
        home_avail = True
        for j in range(fleet.shape[0]):
            if i == fleet[j]['home_evse_id']:
                home_avail = False
                break

        if ((home_avail and not away_bank)  or away_bank) and (random.random() <= bank[i]['probability']):
            if soc < bank[i]['max_soc']:
                # DCFC
                if bank[i]['dc'] and (fleet[vid]['dc_max'] > 0):
                    can_connect = False
                    if (fleet[vid]['dc_plug'] == DCPLUG_CHADEMO) and bank[i]['plug_chademo']:
                        can_connect = True
                    elif (fleet[vid]['dc_plug'] == DCPLUG_COMBO) and bank[i]['plug_combo']:
                        can_connect = True
                    elif (fleet[vid]['dc_plug'] == DCPLUG_TESLA) and bank[i]['plug_tesla']:
                        can_connect = True
                    
                    if can_connect:
                        connected_power = min(fleet[vid]['dc_max'], bank[i]['power_max'])
                        bank[i]['power'] = connected_power
                        if away_bank:
                            fleet[vid]['away_evse_id'] = i
                        else:
                            fleet[vid]['home_evse_id'] = i

                        break
                # AC
                else:
                    connected_power = min(fleet[vid]['ac_max'], bank[i]['power_max'])
                    bank[i]['power'] = connected_power
                    if away_bank:
                        fleet[vid]['away_evse_id'] = i
                    else:
                        fleet[vid]['home_evse_id'] = i
                    break
    
    return bool(connected_power > 0)

@nb.njit(cache=True)
def disconnect_evse(vid, fleet, bank, away_bank = False):
    eid = -1
    if away_bank:
        eid = fleet[vid]['away_evse_id']
    else:
        eid = fleet[vid]['home_evse_id']
    if eid >= 0:
        if not away_bank:
            bank[nb.uint64(eid)]['power'] = 0.
        if away_bank:
            fleet[vid]['away_evse_id'] = -1
        else:
            fleet[vid]['home_evse_id'] = -1          
    return eid

@nb.njit(cache=True)
def disconnect_completed(idx, battery_state, fleet, bank, away_bank = False):
    for vid in range(fleet.shape[0]):
        if idx > 0:
            nrg = battery_state[idx - 1, vid]
            eid = -1

            if away_bank:
                eid = fleet[vid]['away_evse_id']
            else:
                eid = fleet[vid]['home_evse_id']
            
            if eid >= 0:

                soc = get_soc(vid, fleet, nrg)
                if soc >= bank[nb.int64(eid)]['max_soc']:
                    disconnect_evse(vid, fleet, bank, away_bank)

@nb.njit
def charge(idx, battery_state, vid, fleet, eid, bank, min_per_interval):
    if nb.int64(eid) >= 0:
        power = bank[nb.int64(eid)]['power']
        potential = (min_per_interval / 60.0) * power
        max_nrg = fleet[vid]['ev_max_batt'] * bank[nb.int64(eid)]['max_soc']
        prev = fleet[vid]['ev_max_batt']
        if idx > 0:
            prev = battery_state[idx - 1, vid]
        nrg = min(prev+potential, max_nrg)
        battery_state[idx, vid] = nrg


@nb.njit(cache=True)
def charge_connected(idx, battery_state, fleet, home_bank, away_bank, min_per_interval):

    for i in range(fleet.shape[0]):
        home_eid = fleet[i]['home_evse_id']
        away_eid = fleet[i]['away_evse_id']
        if home_eid >= 0:
            charge(idx, battery_state, i, fleet, np.int64(home_eid), home_bank, min_per_interval)
        elif away_eid >= 0:
            charge(idx, battery_state, i, fleet, np.int64(away_eid), away_bank, min_per_interval)
    return battery_state

@nb.njit(cache=True)
def find_next_stop(distance, idx, vid):
    stop = distance.shape[0]
    for i in range(idx, distance.shape[0]):
        if distance[i, vid] == 0.:
            stop = i
            break
    return stop

@nb.njit(cache=True)
def find_stop_end(distance, idx, vid):
    stop = distance.shape[0]
    for i in range(idx, distance.shape[0]):
        if not (distance[i, vid] == 0.):
            stop = i
            break
    return stop


@nb.njit(cache=True)
def stop_length_min(distance, idx, vid, min_per_interval):
    i = 0
    while (idx + i < distance.shape[0]) and (distance[idx+i, vid] == 0.):
        i += 1
    return i * min_per_interval

@nb.njit(cache=True)
def defer_next_trip(distance, idx, vid, deferred):
    # Find end of stop, if necessary
    i = idx
    end = distance.shape[0]
    while (i < end) and (distance[i, vid] == 0.):
        i += 1
    while (i < end) and (not (distance[i, vid] == 0.)):
        deferred[i, vid] = distance[i, vid]
        distance[i, vid] = 0.
        i += 1



@nb.njit(cache=True)
def defer_to_target(distance, deferred):
    for i in range(distance.shape[0]):
        if distance[i] >= 0:
            deferred[i] += distance[i]
    distance[:] = 0.



@nb.njit(cache=True)
def battery_req(distance, idle_load_kw, min_per_interval, vid, fleet):
    nrg = 0
    for idx in range(distance.shape[0]):
        # driving
        if distance[idx, vid] > 0:
            eff = fleet[vid]['ev_eff']
            if eff > 0:
                nrg += distance[idx, vid] / eff
        # idle
        elif distance[idx, vid] < 0:
            nrg += idle_load_kw * (min_per_interval / 60.0)
        idx += 1
    return nrg

@nb.njit(cache=True)
def try_defer_trips(
    idx,
    vid,
    fleet,
    distance,
    deferred,
    mask,
    battery_nrg,
    min_per_interval,
    idle_load_kw,
    away_thresh,
    home_thresh,
    soc_buffer):

    # Only bother with BEVs
    if fleet[vid]['type'] == BEV:
        req = battery_nrg + 100
        nrg = battery_nrg * (1. - soc_buffer)
        end = distance.shape[0]
        stop = end
        i = idx
        while (i < end) and (nrg < req):

            stop = find_next_stop(distance, i, vid)
            stop_mins = stop_length_min(distance, stop, vid, min_per_interval)

            # We have an eligible stop!
            if (stop_mins >= away_thresh) or (stop_mins >= home_thresh) or mask[i]:
                req = battery_req(distance[i:stop,:], idle_load_kw, min_per_interval, vid, fleet)

                if nrg >= req:
                    break
                
                defer_to_target(distance[idx:stop, vid], deferred[idx:stop, vid])

            i = stop
            while distance[i, vid] == 0.:
                i += 1


@nb.njit(cache=True)
def drive(idx, distance, battery_state, fuel_use, fleet, idle_load_kw, min_per_interval):

    if idx < distance.shape[0]:
        for vid in range(fleet.shape[0]):

            ev_eff = fleet[vid]['ev_eff']
            ice_eff = fleet[vid]['ice_eff']
            ice_g_kwh = fleet[vid]['ice_gal_kwh']
            batt = 0
            if idx > 0:
                batt = battery_state[idx -1, vid]
            else:
                batt = fleet[vid]['ev_max_batt']


            batt_used = 0.
            fuel_used = 0.

            # driving
            if distance[idx, vid] > 0:
                d = distance[idx, vid]
                
                # Handle EV
                if (fleet[vid]['type'] in (BEV, PHEV)) and ev_eff > 0:
                    nrg_req = d / ev_eff
                    batt_used = min(batt, nrg_req)
                    d = d - (batt_used * ev_eff)
                
                if  (fleet[vid]['type'] in (ICEV, PHEV)) and ice_eff > 0:
                    fuel_used = d / ice_eff
            # idling
            elif distance[idx, vid] < 0:
                e = idle_load_kw * (min_per_interval / 60.)
                 # Handle EV
                if (fleet[vid]['type'] in (BEV, PHEV)) and ev_eff > 0:
                    batt_used = min(batt, e)
                    e = e - batt_used
                
                if  (fleet[vid]['type'] in (ICEV, PHEV)) and ice_g_kwh > 0:
                    fuel_used = e * ice_g_kwh
            

            battery_state[idx, vid] = batt - batt_used    
            fuel_use[idx, vid] = fuel_used

@nb.njit(cache=True)
def num_occupied_evse(fleet):
    n = 0
    for vid in range(fleet.shape[0]):
        eid = fleet[vid]['home_evse_id']
        if not (np.isnan(eid)):
            n += 1
    return n
        

@nb.njit(cache=True)
def evse_usage(idx, fleet, bank, demand, energy, occupancy, utilization, min_per_interval):

    max_power = bank[0]['bank_power_max']
    num_evse = bank.shape[0]
    total_power = 0.
    total_occupied = 0.
    
    for vid in range(fleet.shape[0]):
        eid = fleet[vid]['home_evse_id']
        if eid >= 0:
            power = bank[eid]['power']
            total_power += power
            total_occupied += 1
            energy[idx, eid] = power * (min_per_interval / 60.)
            demand[idx, eid] = power
    if num_evse > 0:
        occupancy[idx] = total_occupied / num_evse
    else:
        occupancy[idx] = 0
    if max_power > 0:
        utilization[idx] = total_power / max_power
    else:
        utilization[idx] = 0
    


@nb.njit(cache=True)
def simulation_loop(
    distance: np.array,
    fleet: np.array,
    home_bank: np.array,
    away_bank: np.array,
    home_mask: np.array,
    interval_min: float,
    home_thresh_min: float,
    away_thresh_min: float,
    idle_load_kw: float,
    queue_soc: bool = False,
    soc_buffer: float = 0.2):

    nrows = distance.shape[0]
    nvehicles = fleet.shape[0]
    dshape = distance.shape
    nevse = home_bank.shape[0]

    fuel_use = np.zeros(dshape, dtype=np.float32)
    battery_state = np.zeros(dshape, dtype=np.float32)
    elec_demand = np.zeros((nrows, nevse), dtype=np.float32)
    elec_energy = np.zeros((nrows, nevse), dtype=np.float32)
    occupancy = np.zeros(nrows, dtype=np.float32)
    utilization = np.zeros(nrows, dtype=np.float32)
    deferred = np.zeros(dshape, dtype=np.float32)

    queue = np.full(nvehicles, np.nan, dtype=np.float32)

    for idx in range(nrows):

        # Disconnect and dequeue departing vehicles
        for vid in range(nvehicles):
            if not (distance[idx, vid] == 0.):
                bank_dequeue(queue, vid)
                disconnect_evse(vid, fleet, home_bank, False)
                disconnect_evse(vid, fleet, away_bank, True)
        
        # Disconnect completed vehicles
        if idx > 0:
            disconnect_completed(idx, battery_state, fleet, home_bank, False)
            disconnect_completed(idx, battery_state, fleet, away_bank, True)
        
        # Do drive and stop stuff
        if idx > 0 and idx < nrows:

            for vid in range(nvehicles):
            
                # Stop stuff
                if distance[idx, vid] == 0.:

                    # New stop
                    if not (distance[idx -1, vid] == 0.):
                        stop_time = stop_length_min(distance, idx, vid, interval_min)
                        soc = get_soc(vid, fleet, battery_state[idx - 1, vid])
                        
                        # Home stop
                        if stop_time >= home_thresh_min or home_mask[idx]:
                            
                            bank_enqueue(idx, vid, soc, fleet, queue, queue_soc)
                        
                        # Away stop
                        elif stop_time >= away_thresh_min:
                            connect_evse(vid, soc, fleet, away_bank, True)
                    
                    # About to depart
                    elif not (distance[idx + 1, vid] == 0.):
                        try_defer_trips(
                            idx,
                            vid,
                            fleet,
                            distance,
                            deferred,
                            home_mask,
                            battery_state[idx - 1, vid],
                            interval_min,
                            idle_load_kw,
                            away_thresh_min,
                            home_thresh_min,
                            soc_buffer
                        )
                    
        # Process queues and charge
        home_occupied = num_occupied_evse(fleet)
        while home_occupied < nevse:

            vid = pop_low_score(queue)

            if np.isnan(vid):
                break
            
            soc = get_soc(vid, fleet, battery_state[idx - 1, vid])
            
            connect_evse(vid, soc, fleet, home_bank, False )

            home_occupied += 1
        
        # Charge connected vehicles
        charge_connected(idx, battery_state, fleet, home_bank, away_bank, interval_min)

        evse_usage(idx, fleet, home_bank, elec_demand, elec_energy, occupancy, utilization, interval_min)

        drive(idx, distance, battery_state, fuel_use, fleet, idle_load_kw, interval_min)
    
    return (fuel_use, battery_state, deferred, elec_demand, elec_energy, occupancy, utilization)




def run_simulation(ds: DatasetInfo, sc: Scenario, storage_info: StorageInfo):
    """Runs a simulation with a given scenario by
    downloading data, running the simulation, and uploading results.

    Args:
        sc (Scenario): The scenario structure to use.
    
    Returns:
        A `SimulationResult` with all the relevant energy data.
    """
    st = DataHandler(storage_info)

    try:
        # First pull down the data
        
        df = st.read_data(ds.obj_path)

        df_index = df.index

        rows = len(df_index)

        interval_len = df.reset_index()['index'][:2].diff()[1].seconds / 60.0

        # Basic rules creation for now

        mask = pd.Series(
            np.zeros(df.values.shape[0], dtype=np.bool_),
            index=df.index
        )

        for mask_rule in sc.home_mask_rules:
            if mask_rule.get('type') == 'time':
                begin = mask_rule.get('begin', '23:55')
                end = mask_rule.get('end', '00:00')
                mask[df.between_time(begin, end).index] = True

        # Create the fleet
        distance, fleet = fleet_from_df(df, sc.powertrains, sc.distribution)

        fleet_size = fleet.shape[0]

        home_banks = make_evse_banks(sc.home_banks, fleet_size)

        away_banks = make_evse_banks(sc.away_banks, fleet_size)

        num_banks = home_banks.shape[0]

        use_soc_queue = False

        output = simulation_loop(
            distance,
            fleet,
            home_banks,
            away_banks,
            mask.values,
            interval_len,
            sc.home_threshold_min,
            sc.away_threshold_min,
            sc.idle_load_kw,
            use_soc_queue,
            sc.soc_deferment_buffer
        )

        evse_names = []
        for i in range(home_banks.shape[0]):
            if home_banks[i]['dc']:
                evse_names.append('ac{}-{}kW'.format(i, home_banks[i]['dc_max']))
            else:
                evse_names.append('ac{}-{}kW'.format(i, home_banks[i]['ac_max']))

        fuel_use, battery_state, deferred, elec_demand, elec_energy, occupancy, utilization = output

        vehicle_ids = df.columns.values.tolist()
        dfcols = df.columns

        fuel_df = pd.DataFrame(fuel_use, columns=dfcols, index=df_index)
        demand_df = pd.DataFrame(elec_demand, columns=evse_names, index=df_index)
        energy_df = pd.DataFrame(elec_energy, columns=evse_names, index=df_index)
        occupancy_df = pd.DataFrame({'occupancy': occupancy, 'utilization': utilization}, index=df_index)
        deferred_df = pd.DataFrame(deferred, columns=dfcols, index=df_index)
        battery_df = pd.DataFrame(battery_state, columns=dfcols, index=df_index)
        
        obj_base = 'results/' + 'simulations/' + sc.run_id + '/'

        dfs = [fuel_df, demand_df, battery_df, occupancy_df, deferred_df, energy_df]
        lbls = ['fuel', 'demand', 'battery', 'occupancy', 'deferred', 'energy']

        results = []

        for fr, nm in zip(dfs, lbls):
            
            result = st.upload_data(
                df=fr,
                obj_path=obj_base + nm + '/' +  ds.dataset_id,
                formats='json records csv',
            )
            if result:
                results.extend(result)

        
        

    except Exception as e:
        raise e

    finally:
        st.cleanup()

    return results






        
            









import numpy as np
import pandas as pd

from .vehicle import Vehicle
from .powertrain import Powertrain
from .eligibility import EligibilityRules

__all__ = [
    'Fleet'
]

class Fleet(object):
    """A collection of vehicles and data connecting them."""

    __slots__ = (
        'df',
        'vehicles',
        'size'
    )

    def __init__(self, df: pd.DataFrame, powertrains: list,
                 probabilities: list, rules: EligibilityRules):
        self.df = df.astype(np.float32, copy=False)
        self.vehicles = []
        self.size = self.df.values.shape[1]

        vnames = df.columns

        if isinstance(powertrains, list):
            for pt in powertrains:
                if not isinstance(pt, Powertrain):
                    raise TypeError()

            trains = powertrains
            num_trains = len(powertrains)
            num_probs = len(probabilities)
            probs = probabilities
            if not probabilities:
                probs = [float(1.0 / num_trains) for i in range(num_trains)]
            elif num_probs < num_trains:
                current_sum = sum(probs)
                if current_sum < 1:
                    remaining = num_trains - num_probs
                    distribution = (1 - current_sum) / remaining
                    probs.extend([distribution for i in range(remaining)])
                else:
                    trains = powertrains[:num_probs]
            elif num_trains < num_probs:
                probs = probabilities[:num_trains]

            pt_assignments = np.random.choice(
                range(num_trains),
                size=self.size,
                replace=True,
                p=np.array(probs)
            )

            for pt, vid, vname in zip(pt_assignments, range(self.size), vnames):
                vehicle = Vehicle(
                    vid,
                    trains[pt],
                    self.df.values[:, vid],
                    rules,
                    vid = vname
                )
                self.vehicles.append(vehicle)

    @property
    def vehicle_ids(self):
        return self.df.columns.to_list()

    @property
    def fleet_ids(self):
        return [self.df.columns.get_loc(i) for i in self.df.columns]

    def add_vehicle(self, col_id, pt: Powertrain, rules: EligibilityRules):
        if not (col_id in self.vehicle_ids):
            raise IndexError()

        v = Vehicle(
            list(self.df.columns).index[col_id],
            self.df[col_id].values,
            rules
        )
        self.vehicles.append(v)

    def assign_powertrains(self, powertrains: list, 
                           probabilities: list, rules: EligibilityRules):

        self.vehicles = []

        if isinstance(powertrains, list):
            for pt in powertrains:
                if not isinstance(pt, Powertrain):
                    raise TypeError()

            trains = powertrains
            num_trains = len(powertrains)
            num_probs = len(probabilities)
            probs = probabilities
            if not probabilities:
                probs = [float(1.0 / num_trains) for i in range(num_trains)]
            elif num_probs < num_trains:
                current_sum = sum(probs)
                if current_sum < 1:
                    remaining = num_trains - num_probs
                    distribution = (1 - current_sum) / remaining
                    probs.extend([distribution for i in range(remaining)])
                else:
                    trains = powertrains[:num_probs]
            elif num_trains < num_probs:
                probs = probabilities[:num_trains]

            pt_assignments = np.random.choice(trains, self.size, True, probs)

            for pt, vid in zip(pt_assignments, self.vehicle_ids):
                self.add_vehicle(vid, pt)

    def get_powertrain_distribution(self, powertrains: list,
                                    probabilities: list,
                                    rules: EligibilityRules):

        self.vehicles = []

        if isinstance(powertrains, list):
            for pt in powertrains:
                if not isinstance(pt, Powertrain):
                    raise TypeError()

            trains = powertrains
            num_trains = len(powertrains)
            num_probs = len(probabilities)
            probs = probabilities
            if not probabilities:
                probs = [float(1.0 / num_trains) for i in range(num_trains)]
            elif num_probs < num_trains:
                current_sum = sum(probs)
                if current_sum < 1:
                    remaining = num_trains - num_probs
                    distribution = (1 - current_sum) / remaining
                    probs.extend([distribution for i in range(remaining)])
                else:
                    trains = powertrains[:num_probs]
            elif num_trains < num_probs:
                probs = probabilities[:num_trains]        

            pt_assignments = np.random.choice(trains, self.size, True, probs)

            for pt, vid in zip(pt_assignments, self.vehicle_ids):
                self.add_vehicle(vid, pt)



        
        
        



import numpy as np
import math

from typing import NamedTuple, List
import uuid

from .powertrain import Powertrain
from .evse import EVSEType, EVSE
from .plug import DCPlug
from .bank import QueueMode, Bank


class Scenario(NamedTuple):
    powertrains: List[Powertrain]
    distribution: list
    run_id: str = uuid.uuid4().hex
    home_threshold_min: float = 180.0 # Default to 3 hours
    away_threshold_min: float = 90.0 # default to 1.5 hours
    soc_deferment_buffer: float = .2 # default to 20% SoC buffer
    idle_load_kw: float =  0.
    home_mask_rules: list = [
        # An example of a time-based mask
        {
            'type': 'time',
            'begin': '19:00',
            'end': '08:00'
        }
        # You could feasibly also use a location-based
        #{
            # Type
        #    'type': 'location',
            # Some lat
        #    'latitude': 12.5987653,
            # some long
        #    'longitude': 13.908643,
            # Some radius in meters
        #    'radius': 400
        #}
    ]
    home_banks: list = [
        {
            'probability': 1.,
            'queue': QueueMode.DEFAULT,
            'evse': [
                (EVSEType(7), 0, 0, .5)
            ]
        }
    ]
    away_banks: list = [
        {
            'probability': .1,
            'evse': [
                (EVSEType(50, True, .8, (DCPlug.CHADEMO, DCPlug.COMBO)), 1, 0, 0)
            ]
        },
        {
            'probability': .2,
            'evse': [
                (EVSEType(7.0), 1, 0, 0)
            ]
        }
    ]

    def make_home_banks(self, fleet_size, rows):
        home_banks = []
        for bank_info in self.home_banks:
            evse_list = []
            for evse_type, nmin, nmax, proportion in bank_info.get('evse'):
                n_evse = min(nmin, max(nmax, math.floor(proportion * fleet_size)))
                for i in range(n_evse):
                    evse_list.append(EVSE(evse_type))
            bank = Bank(
                max_power=bank_info.get('max_power', 0.),
                capacity=bank_info.get('capacity', 0.),
                evse=evse_list,
                queue_probability=bank_info.get('probability', 1.),
                queue_mode=bank_info.get('queue', QueueMode.DEFAULT),
                demand_profile=np.zeros(rows, dtype=np.float32),
                occupancy_profile=np.zeros(rows, dtype=np.uint8),
                dynamic_size=False
            )
            home_banks.append(bank)
        return home_banks
    
    def make_away_banks(self, rows):
        away_banks = []
        for bank_info in self.away_banks:
            evse_list = []
            for evse_type in bank_info.get('evse'):
                evse_list.append(EVSE(evse_type))
            demand_a = np.zeros(rows, dtype=np.float32)
            occupancy_a = np.zeros(rows, dtype=np.uint8)
            bank = Bank(
                max_power=bank_info.get('max_power', 0.),
                capacity=bank_info.get('capacity', 0.),
                evse=evse_list,
                queue_probability=bank_info.get('probability', .2),
                queue_mode=bank_info.get('queue', QueueMode.DEFAULT),
                demand_profile=np.zeros(rows, dtype=np.float32),
                occupancy_profile=np.zeros(rows, dtype=np.uint8),
                dynamic_size=True
            )
            away_banks.append(bank)
        return away_banks




from dask.distributed import progress
import dask
import os
import uuid
from typing import List
import math
import logging

from dask_k8 import DaskCluster

from .scenario import Scenario
from .dataset import DatasetInfo
from .datastorage import StorageInfo
from .simulation import run_simulation, SimulationResult
from .job_results import JobResults

from requirements import requirements

DEFAULT_DASK_SCHEDULER_YAML = """
containers:
  - image: daskdev/dask:latest
    command: ["sh", "-c"]
    args:
      - dask-scheduler --port 8786 --bokeh-port 8787
    imagePullPolicy: Always
    name: dask-scheduler
    ports:
      - containerPort: 8787
      - containerPort: 8786
    resources:
      requests:
          cpu: 1
          memory: 4G
"""

DEFAULT_DASK_WORKER_YAML = """
containers:
  - image: daskdev/dask:latest
    args: [dask-worker, $(DASK_SCHEDULER_ADDRESS), --nthreads, '1', --no-bokeh, --memory-limit, 3GB, --death-timeout, '60']
    imagePullPolicy: Always
    name: dask-worker
    env:
      - name: POD_IP
        valueFrom:
        fieldRef:
          fieldPath: status.podIP
      - name: POD_NAME
        valueFrom:
        fieldRef:
          fieldPath: metadata.name
      - name: EXTRA_PIP_PACKAGES
        value: 
      - name: EXTRA_CONDA_PACKAGES
        value: {conda_pkgs}
    resources:
      requests:
        cpu: 1
        memory: "3G"
      limits:
        cpu: 1
        memory: "3G"
""".format(
    conda_pkgs=' '.join(r for r in requirements)
)

class KubeJobCollection(object):
    """Runs a collection of scenarios and datasets through the simulation using Dask
    and Kubernetes.
    """

    __slots__ = (
        'storage_info',
        'sched_yaml',
        'worker_yaml',
        'namespace',
        'cluster_id',
        'n_workers',
        'client'
    )

    def __init__(self, storage_info: StorageInfo, num_workers: int,
                 namespace: str, cluster_id: str = 'evnrg-dask-{}'.format(uuid.uuid4().hex),
                 scheduler_pod_file: str = None, worker_pod_file: str = None,
                 scheduler_pod_spec: str = None, worker_pod_spec: str = None):

      if num_workers <= 0:
          raise ValueError('Number of workers must be >= 0!')
      
      # Initialize lists of stuff
      self.storage_info = storage_info,
      self.n_workers = num_workers
      
      self.sched_yaml = DEFAULT_DASK_SCHEDULER_YAML
      self.worker_yaml = DEFAULT_DASK_WORKER_YAML
      self.cluster_id = cluster_id
      self.namespace = namespace
      self.client = None

      if isinstance(scheduler_pod_file, str) and scheduler_pod_file:
          try:
              shed_file = open(scheduler_pod_file, 'r')
              self.sched_yaml = sched_file.read()
          except IOError:
              raise IOError('Schedule pod YAML file does not exist.')
      if isinstance(worker_pod_file, str) and worker_pod_file:
          try:
              work_file = open(worker_pod_file, 'r')
              self.worker_yaml = work_file.read()
          except IOError:
              raise IOError('Schedule pod YAML file does not exist.')
      
      if scheduler_pod_spec and isinstance(scheduler_pod_spec, str):
          self.sched_yaml = scheduler_pod_spec
      if worker_pod_spec and isinstance(worker_pod_spec, str):
          self.worker_yaml = worker_pod_spec
        
    def run_simulations(self, scenarios: List[Scenario], datasets: List[DatasetInfo], print_client_info = True):

      logging.info('Creating cluster object...')
      cluster = DaskCluster(
          namespace=self.namespace,
          cluster_id=self.cluster_id,
          worker_pod_spec=self.worker_yaml,
          scheduler_pod_spec=self.sched_yaml
      )
      out = []

      logging.info('Starting cluster...')
      with cluster:
          logging.info('Making Dask client...')
          client = cluster.make_dask_client()  # Waits for the scheduler to be started
          if print_client_info:
            print(client)
          cluster.scale(self.n_workers)

          results = []

          logging.info('Running Scenarios...')
          for scenario in scenarios:
              for dataset in datasets:
                  sim_result = client.submit(run_simulation, dataset, scenario, self.storage_info)
                  results.append(sim_result)
          
          out = results.result()
      
      return JobResults(out)
    
    @classmethod
    def calc_num_workers(cls, worker_mem: float, worker_cpu: float,
                       sched_mem: float, sched_cpu: float,
                       total_mem: float, total_cpu: float,
                       max_util: float = .85):
      
      mem_max = math.floor(((total_mem - sched_mem) * max_util) / worker_mem)
      cpu_max = math.floor(((total_cpu - sched_cpu) * max_util) / worker_cpu)

      return min(mem_max, cpu_max)




from typing import NamedTuple

import numpy as np
import numba as nb

from .eligibility import EligibilityRules
from .eligibility import check_mask
from .status import Status


@nb.jitclass([
    ('index', nb.int64),
    ('begin', nb.int64),
    ('status', nb.uint8),
    ('index_distance', nb.float32),
    ('total_distance', nb.float32)
])
class Bracket(object):
    """A jitclass that provides information about the current drive or stop.

    Attributes:
        index (numba.int64): The index in the distance array that the bracket
            was created.
        begin (numba.int64): The beginning of the bracket
        end (numba.int64): The end of the bracket. Also indicates the following
            bracket's `begin`.
        status (numba.uint8): The current 'Eligibility` for the bracket.
        index_distance (numba.float32): The current distance at the index (km).
        total_distance (numba.float32): The total distance traveled in the
            bracket (km).

    Args:
        idx (int): The current index
        vidx (int): The fleet index (column index) of the vehicle.
        distance_array (numpy.array): A 1-D distance interval array.
        rules (EligibilityRules): The EligibilityRules object that
            determines charging eligibility.
    """

    def __init__(self, idx, vidx, distance_array, rules):
        """Creates a Bracket object"""

        self.index = idx
        self.end = idx
        self.begin = idx
        self.total_distance = 0.
        self.index_distance = 0.
        self.status = Status.NONE

        max_len = distance_array.shape[0]

        driving = False
        if distance_array[idx] > 0.:
            driving = True

        # Only do the following if idx is inside the array.
        # If it is equal or greater than the array length,
        # we have a Bracket that represents the end of the array
        if idx < max_len:

            # Find the beginning
            while self.begin > 0:
                if driving:
                    if not (distance_array[self.begin - 1] > 0.):
                        break
                else:
                    if distance_array[self.begin - 1] > 0.:
                        break
                self.begin -= 1

            # Find the end
            while self.end < max_len:
                if driving:
                    if not (distance_array[self.end] > 0.):
                        break
                else:
                    if distance_array[self.end] > 0.:
                        break
                self.end += 1

            self.index_distance = distance_array[idx]

            length = self.end - self.begin
            # See if we've been stopped long enough to trigger
            # a charge opportunity
            home = False
            if rules.home_threshold > 0 and length > rules.home_threshold:
                home = True

            away = False
            if rules.away_threshold > 0 and length > rules.away_threshold:
                away = True

            if home and away:
                self.status = rules.threshold_priority
            elif home:
                self.status = Status.HOME_ELIGIBLE
            elif away:
                self.status = Status.AWAY_ELIGIBLE

            # Masks override any time-based rules
            # This allows for time-fencing and geo-fencing at the data
            # preparation level.
            charge_mask = check_mask(self.begin, self.end, vidx, rules)

            if charge_mask in {Status.HOME_ELIGIBLE, Status.AWAY_ELIGIBLE}:
                self.status = charge_mask

            self.total_distance = distance_array[self.begin:self.end].sum()
        # Endif

    @property
    def length(self):
        """Returns the number of intervals in the bracket"""
        return self.end - self.begin

    @property
    def driving(self):
        """Returns if the bracked is a driving one."""
        return self.index_distance > 0.

    @property
    def stopped(self):
        """Returns if the bracked is a stop."""
        return not self.is_driving


# Deprecated the NamedTuple implementation
class BracketNT(NamedTuple):
    """Deprecated, but kept for reference"""
    begin: int = -1
    end: int = -1
    index: int = -1
    length: int = 0
    mask: bool = False
    status: Status = Status.STOPPED
    dist: float = 0.
    idist: float = 0.
    stopped: bool = True


def update_bracket(bkt: BracketNT, d: dict):
    """Deprecated"""
    x = bkt._todict
    _d = {k: d[k] for k, v in Bracket.items() if k in Bracket._fields}
    x.update(_d)

    return Bracket(**x)


@nb.njit
def make_bracket(arr: np.array, index: int, einfo: EligibilityRules):
    """Deprecated"""

    alen = arr.shape[0]
    mask = einfo.home_mask

    assert alen == mask.shape[0], \
        'Mask and distance array lengths do not match.'

    e = Status.NONE
    idist = arr[index]
    stopped = True
    # If we're not currently stopped, don't bother.
    if idist > 0:
        e = Status.DRIVING
        stopped = False

    begin = index + 0
    end = index + 0

    def _test_(a, b):
        x = a > 0.  # Test for driving
        if b:  # If stopped
            x = not (a > 0.)  # Then invert test
        return x

    # search backwards
    while begin > 0:
        if _test_(arr[begin - 1], stopped):
            break
        begin -= 1

    # Search forwards
    while end < alen:
        if _test_(arr[end], stopped):
            break
        end += 1

    home_mask_match = (mask[begin:end].sum() > 0)
    dist = arr[begin:end].sum()
    length = end - begin

    if home_mask_match:
        e = Status.HOME_ELIGIBLE

    elif stopped:
        for p in einfo.priority:

            if p == Status.HOME_ELIGIBLE:
                if length >= einfo.home_threshold:
                    e = Status.HOME_ELIGIBLE

            elif p == Status.AWAY_ELIGIBLE:
                if length >= einfo.away_threshold:
                    e = Status.AWAY_ELIGIBLE

    return BracketNT(
        begin=begin,
        end=end,
        index=index,
        length=length,
        mask=home_mask_match,
        status=e,
        dist=dist,
        idist=idist,
        stopped=stopped
    )
    


from typing import NamedTuple

class JobResults(NamedTuple):

    results: list

    @property
    def failed(self):
        failed = []
        for r in results:
            if r.error:
                failed.append(r)
        return failed
    
    @property
    def succeeded(self):
        good = []
        for r in results:
            if not r.error:
                good.append(r)
        return good


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

from evnrg import __version__
# from requirements import requirements, dependency_links

requirements = [
    'numba>=0.42.0',
    'numpy>=1.16.0',
    'pandas>=0.24.0',
    'dask',
    'distributed',
    'appdirs',
    'apache-libcloud',
    'fastparquet',
    'pycrypto'
]

dependency_links = []

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', ]

setup(
    author="Alex Campbell",
    author_email='amcampbell@ucdavis.edu',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="EVNRG is an EV electrical demand simulation package that takes in trip data and turns it into useful energy data given a set of assumptions.",
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='evnrg',
    name='evnrg',
    packages=find_packages(include=['evnrg']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/accurrently/evnrg',
    version=__version__,
    zip_safe=False,
    dependency_links=dependency_links
)



from typing import List

from .eligibility import EligibilityRules
from .status import Status
from .bank import Bank
from .vehicle import Vehicle


__all__ = []



def simulation_loop(
        vehicles: List[Vehicle],
        home_banks: List[Bank],
        away_banks: List[Bank],
        rules: EligibilityRules,
        min_per_interval: float,
        num_intervals: int,
        idle_load_kw: float = 0.
        ):

    for index in range(num_intervals):

        # Rewrite any trips that might not be possible
        for vehicle in vehicles:
            vehicle: Vehicle
            vehicle.attempt_defer_trips(rules, min_per_interval, idle_load_kw)
        
        # Disconnect all the vehicles that are leaving or done
        # Home banks
        for bank in home_banks:

            bank.dequeue_early_departures(index)
            bank.disconnect_completed_vehicles(index)
            bank.disconnect_departing_vehicles(index)
        # Away banks
        for bank in away_banks:
            bank.disconnect_completed_vehicles(index)
            bank.disconnect_completed_vehicles(index)
            bank.disconnect_departing_vehicles(index)            

        # Drive vehicles or enter queues
        for vehicle in vehicles:
            if index >= vehicle.idx:
                
                if vehicle.status == Status.HOME_ELIGIBLE:
                    for bank in home_banks:
                        if bank.enqueue_vehicle_prob(vehicle):
                            break
                elif vehicle.status == Status.AWAY_ELIGIBLE:
                    for bank in away_banks:
                        if bank.enqueue_vehicle_prob(vehicle):
                            break

        # Process the EVSE queue and charge
        for bank in home_banks:
            bank.process_queue()
            bank.charge_connected(min_per_interval, index)

        for bank in away_banks:
            bank.process_queue()
            bank.charge_connected(min_per_interval, index)

        for vehicle in vehicles:
            if index >= vehicle.idx:
                vehicle.advance_index(rules, min_per_interval, idle_load_kw)


    return vehicles, home_banks



import enum
import random
import math
import copy
from typing import List

import numpy as np

from .evse import EVSE
from .vehicle import Vehicle
from .status import Status

__all__ = [
    'QueueMode',
    'Bank'
]


class QueueMode(enum.IntEnum):

    DEFAULT = enum.auto()
    STACK = enum.auto()  # Last-in, first out
    RANDOM = enum.auto()
    SOC = enum.auto()
    TIME = enum.auto()
    RSOC = enum.auto()
    RTIME = enum.auto()

    @classmethod
    def lookup(cls, s: str):
        return {
            'default': QueueMode.DEFAULT,
            'stack': QueueMode.STACK,
            'random': QueueMode.RANDOM,
            'soc': QueueMode.SOC,
            'time': QueueMode.TIME,
            'rsoc': QueueMode.RSOC,
            'rtime': QueueMode.RTIME
        }.get(s, QueueMode.DEFAULT)


ALLOWED_QUEUE_MODES = {
    QueueMode.DEFAULT,
    QueueMode.STACK,
    QueueMode.RANDOM,
    QueueMode.SOC,
    QueueMode.TIME,
    QueueMode.RSOC,
    QueueMode.RTIME
}


class Bank(object):

    __slots__ = (
        'max_power',
        'capacity',
        'queue_probability',
        'dynamic_size',
        'queue_mode',
        'queue',
        'available_evse',
        'occupied_evse',
        'demand_profile',
        'occupancy_profile'
    )

    def __init__(self, max_power: float = 0., capacity: float = 0., queue_probability: float = 1.,
                 dynamic_size: bool = False, queue_mode: QueueMode = QueueMode.DEFAULT,
                 evse: list = [], demand_profile: np.array = np.empty(0),
                 occupancy_profile: np.array = np.empty(0)):
        
        self.max_power = max_power
        self.capacity = capacity
        self.queue_probability = queue_probability
        self.dynamic_size = dynamic_size
        self.queue_mode = queue_mode
        self.queue = []
        self.available_evse = evse
        self.occupied_evse = []
        self.demand_profile = demand_profile
        self.occupancy_profile = occupancy_profile

    @property
    def num_available(self):
        return len(self.available_evse)

    @property
    def occupancy(self) -> float:
        if self.size > 0:
            return float(len(self.occupied_evse) / self.size)
        return 0

    @property
    def num_occupied(self):
        return len(self.occupied_evse)
    
    @property
    def size(self):
        return self.num_available + self.num_occupied

    @property
    def num_operating(self) -> int:
        x = 0
        for evse in self.occupied_evse:
            if evse.is_connected:
                x += 1
        return x

    @property
    def total_demand(self) -> float:
        x = 0.0
        for evse in self.occupied_evse:
            if evse.is_connected:
                x += evse.demand
        return x

    @property
    def pct_capacity(self):
        if self.capacity > 0:
            return self.total_demand / self.capacity
        return 0

    def record_demand(self, idx: int):
        self.demand_profile[idx] = self.total_demand
    
    def record_occupancy(self, idx: int):
        self.occupancy_profile[idx] = self.num_occupied

    def add_evse_with_rules(self, fleet_size: int,
                            chargers: List[EVSE], rules: dict):
        for evse, rule in zip(chargers, rules):
            min_num = rule.get('minimum', 1)
            max_num = rule.get('maximum', 0)
            ratio = rule.get('evse_per_vehicle', 0)

            num_to_add = max(min_num, fleet_size*ratio)

            if max_num > 0:
                num_to_add = min(num_to_add, max_num)

            num_to_add = math.ceil(num_to_add)

            for i in range(num_to_add):
                self.available_evse.append(copy.deepcopy(evse))
                if self.dynamic_size:
                    break

            if self.dynamic_size:
                break

    def charge_connected(self, minutes_per_interval: float, idx: int):
        demand = 0.
        occ = 0
        for evse in self.occupied_evse:
            if evse.vehicle.idx <= idx:
                demand += evse.charge_vehicle(minutes_per_interval)
                occ += 1
        self.demand_profile[idx] = demand
        self.occupancy_profile[idx] = (occ / self.size) if self.size > 0 else 0

    def set_profile_length(self, size: int):
        self.demand_profile = np.zeros(size, dtype=np.float32)
        self.occupancy_profile = np.zeros(size, dtype=np.uint8)

    def add_evse(self, evse: EVSE):
        self.available_evse.append(evse)
        self.size = min(255, self.size + 1)
        self.capacity += evse.max_power

    def remove_evse(self, evse: EVSE):
        out = True
        try:
            self.capacity = max(0, self.capacity - evse.max_power)
            self.available_evse.remove(evse)
            self.size = max(0, self.size - 1)
        except Exception:
            out = False
        return out

    def enqueue_vehicle(self, vic: Vehicle):
        if vic.powertrain.pev:
            vic.status = Status.IN_QUEUE
            self.queue.append(vic)

    def enqueue_vehicle_prob(self, vic: Vehicle):
        out = False
        if vic.powertrain.pev:
            if random.random() <= self.queue_probability:
                self.enqueue_vehicle(vic)
                out = True
        return out

    def dequeue_vehicle(self, vic: Vehicle):
        out = False
        try:
            self.queue.remove(vic)
            out = True
        except Exception:
            out = False
        return out

    def dequeue_vehicle_index(self, fleet_index: int):
        out = None
        for v in self.queue:
            if v.fleet_index == fleet_index:
                if self.dequeue_vehicle(v):
                    out = v
        return out

    def next_available_evse(self):
        out = None
        if self.available_evse:
            if not self.available_evse[0].is_connected:
                out = self.available_evse[0]
        return out

    def dequeue_early_departures(self, idx: int):
        departing = []
        for v in self.queue:
            if v.idx <= idx:
                if v.is_driving:
                    if self.dequeue_vehicle(v):
                        departing.append(v)
        return departing

    def pop_next_vehicle(self):
        winner = None
        if self.queue:
            # Default queue mode
            winner = self.queue[0]

            # Stack
            if self.queue_mode == QueueMode.STACK:
                winner = self.queue[-1]

            # Random
            elif self.queue_mode == QueueMode.RANDOM:
                winner = self.queue[random.randrange(len(self.queue))]

            # SoC (lowest first)
            elif self.queue_mode == QueueMode.SOC:
                for v in self.queue[1:]:
                    if v.soc < winner.soc:
                        winner = v

            # SoC (Highest first)
            elif self.queue_mode == QueueMode.RSOC:
                for v in self.queue[1:]:
                    if v.soc > winner.soc:
                        winner = v

            # Time to charge (assume 6kW power)
            elif self.queue_mode == QueueMode.TIME:
                for v in self.queue[1:]:
                    v_time = v.minutes_to_charged(6, 1)
                    winner_time = winner.minutes_to_charged(6, 1)
                    if v_time > winner_time:
                        winner = v

            # Lowest time to charge
            elif self.queue_mode == QueueMode.RTIME:
                for v in self.queue[1:]:
                    v_time = v.minutes_to_charged(6, 1)
                    winner_time = winner.minutes_to_charged(6, 1)
                    if v_time < winner_time:
                        winner = v

            if not self.dequeue_vehicle(winner):
                winner = None
        return winner

    def process_queue(self):
        """Checks each EVSE in the bank.
        If the EVSE is not connected to a vehicle, dequeue one and connect it.
        """
        num_connected = 0

        # Only do work if there are available EVSE
        if self.num_available > 0:

            # If we're dynamically sized, conect every Vehicle
            # that is compatible with the type of charger.
            # If a Vehicle is ineligible (incompatible or the SoC is too high),
            # remove it from the queue
            if self.dynamic_size:
                while self.queue:
                    vehicle = self.pop_next_vehicle()
                    evse: EVSE
                    evse = self.available_evse[0]
                    if vehicle.evse_compatible(evse.model.dc, evse.model.dc_plugs):
                        evse = copy.deepcopy(self.available_evse[0])
                        if evse.connect_vehicle(vehicle):
                            self.occupied_evse.append(evse)
                            num_connected += 1

            # Otherwise, process a queue normally
            else:
                ineligible_vehicles = []

                # As long as there is EVSE available and there are still
                # vehicles in the queue
                while self.num_available > 0 and self.queue:
                    vehicle = self.pop_next_vehicle()
                    
                    # Check for a valid vehicle
                    if vehicle is not None:
                        connected = False
                        for evse in self.available_evse:
                            # Check if the EVSE can accept connections
                            # and the EVSE is compatible with the Vehicle and 
                            # its needs
                            if evse.connect_vehicle(vehicle):
                                num_connected += 1      # Update our count

                                # Move the EVSE to the list of occupied
                                self.available_evse.remove(evse)
                                self.occupied_evse.append(evse)

                                # Remember that we connected this one
                                connected = True
                                # Break out of the loop;
                                # No need to look at other EVSE
                                break

                        # If we failed to connect, the vehicle is ineligible
                        # for any of the available EVSE.
                        if not connected:
                            ineligible_vehicles.append(vehicle)

                # Now that we're done popping Vehicles,
                # requeue the ineligible ones so they can wait for EVSE that
                # will meet their needs

                # First, check if we're using a stack, so we can re-stack the
                # queue
                if self.queue_mode == QueueMode.STACK:
                    ineligible_vehicles = ineligible_vehicles[::-1]

                for v in ineligible_vehicles:
                    self.queue.append(v)

        # Return the number of vehicles that got to connect
        return num_connected

    def disconnect_completed_vehicles(self, idx: int):
        """
        Disconnects any vehicles that are effectively done charging.
        """
        num_disconnected = 0
        for evse in self.occupied_evse:
            if evse.is_connected and not evse.v2g_mode:
                vehicle = evse.vehicle
                if vehicle.idx <= idx:
                    if evse.charge_completed:
                        if evse.disconnect_vehicle() is not None:
                            num_disconnected += 1
                            self.occupied_evse.remove(evse)
                            if not self.dynamic_size:
                                self.available_evse.append(evse)
        return num_disconnected

    def disconnect_departing_vehicles(self, idx: int):
        num_disconnected = 0
        for evse in self.occupied_evse:
            if evse.is_connected:
                vehicle = evse.vehicle
                if vehicle.idx <= idx:
                    if evse.vehicle.is_driving:
                        v = evse.vehicle
                        if evse.disconnect_vehicle() is not None:
                            v.status = Status.DRIVING
                            num_disconnected += 1
                            self.occupied_evse.remove(evse)
                            if not self.dynamic_size:
                                self.available_evse.append(evse)
        return num_disconnected




import math
from typing import NamedTuple

from .vehicle import Vehicle
# from evnrg.simulation.eligibility import EligibilityRules

__all__ = [
    'EVSE',
    'EVSEType'
]

class EVSEType(NamedTuple):
    max_power: float = 0.
    dc: bool = False
    max_soc: float = 1.
    dc_plugs: tuple = (None,)
    v2g_buffer: float = 0.20  # Arbitrage of 20% max SoC
    v2g_capable: bool = False


class EVSE(object):

    def __init__(self, model: EVSEType = EVSEType()):
        self.model = model
        self.power = model.max_power
        self.throttle = 1.
        self.vehicle = None
        self.target_soc = model.max_soc
        self.v2g_mode = False

    # TODO: Create properties that show the maximum and minimum SoCs and
    # energy available for V2G use.
    # Can also create methods to describe temporally-based V2G potential
    # For the purpose ofsafely bidding into an ISO/utility's grid services,
    # we could estimate this as:
    # (time_until_departure / 2) * charge_power

    @property
    def is_connected(self):
        return self.vehicle is not None
    
    @property
    def is_available(self):
        return self.vehicle is None

    @property
    def demand(self):
        out = 0
        if self.is_connected:        
            if self.v2g_mode:
                # TODO: If the vehicle is discharging in V2G mode
                pass
            else:
                out = self.power * self.throttle
        return out

    @property
    def charge_completed(self):
        out = False
        if self.vehicle is not None:
            out = self.vehicle.soc >= self.target_soc
        return out

    def charge_vehicle(self, minutes: float):
        true_power = 0
        if self.is_connected:
            if not self.charge_completed:
                power = self.power * self.throttle
                energy_max = self.vehicle.powertrain.energy_at_soc(self.target_soc)
                energy_potential = power * (minutes / 60.0)
                energy_limit = energy_max - self.vehicle.battery_state
                energy_to_add = min(energy_potential, energy_limit)
                self.vehicle.charge_battery(energy_to_add)
                true_power = energy_to_add / (minutes / 60.0)
        return true_power

    def connect_vehicle(self, v: Vehicle):
        out = False
        target = min(self.target_soc, v.max_soc)
        if target > v.soc:
            if v.connect_evse(self.power, self.model.dc, self.model.dc_plugs):
                self.target_soc = target
                power = 0.
                if self.model.dc and v.powertrain.dc_power > 0.:
                    power = min(self.model.max_power, v.powertrain.dc_power)
                else:
                    power = min(self.model.max_power, v.powertrain.ac_power)
                self.power = power
                self.vehicle = v
                out = True
        return out

    def intervals_to_target_soc(self, minutes_per_interval: float):
        out = 0
        if self.vehicle:
            current_energy = self.vehicle.battery
            target_energy = self.vehicle.energy.energy_at_soc(self.target_soc)

            energy_needed = target_energy - current_energy

            hours_required = energy_needed / self.power

            out = math.ceil(hours_required / minutes_per_interval)
        return out

    def connect_vehicle_soc_target(self, v: Vehicle, desired_soc: float):
        self.target_soc = desired_soc
        return self.connect_vehicle(v)

    def disconnect_vehicle(self):
        out = None
        if self.is_connected:
            v = self.vehicle
            self.vehicle.disconnect_evse()
            self.vehicle = None
            self.target_soc = self.model.max_soc
            out = v
        return out
